import calendar
import unittest
from datetime import datetime, date

import psycopg2

import pg_creds
from simpleetl import FactTable, runETL, datatypes, CONFIG

SCHEMA = "testschema"
TABLE = "test_partitioning"
CONFIG.batch_size = 1
CONFIG.parallel_processes = 4
from simpleetl._functions._dbfunctions import get_table_partitions


def datagenerator(rangeadd=0):
    for n in range(24):
        n += rangeadd
        add_years, month = divmod(n, 12)
        month += 1
        year = 2010 + add_years
        yield {"lookupid": n, "datekey": "{y}{m}01".format(y=year, m=str(month).rjust(2, '0'))}


def partitionfunc_yearmonth(value):
    """Function for telling simpleetl to partition facttable by year and month.
    :parameter value: Input is an integer datekey, ex. '2019-11-23' is 20191123
    :return: Tuple with suffix, range_from and range_to, ex. ('201911', 20191101, 20191131)
    """
    dt = datetime.strptime(str(value), "%Y%m%d")
    first_day, last_day = calendar.monthrange(dt.year, dt.month)
    suffix = dt.strftime("%Y_%m")  # 201911
    range_from = int(date(dt.year, dt.month, 1).strftime("%Y%m%d"))  # 20191101, ensures zeropadded month
    range_to = int(date(dt.year, dt.month, last_day).strftime("%Y%m") + str(last_day + 1))  # 20191131
    return suffix, range_from, range_to


def partitionfunc_year(value):
    suffix = str(int(value / 10000))
    range_from = int("{y}0101".format(y=int(value / 10000)))
    range_to = int("{y}0101".format(y=int(value / 10000) + 1))

    return suffix, range_from, range_to


def etlrun(rangeadd):
    f = FactTable(schema=SCHEMA, table=TABLE, lookupatts=["lookupid", "datekey"],
                  store_history=True)
    f.enable_partitioning("datekey", partitionfunc_yearmonth)
    f.enable_partitioning_historic("datekey", partitionfunc_year)
    f.handle_deleted_rows(method="wipe")
    f.add_column_mapping("lookupid", datatypes.bigint)

    f.add_column_mapping("datekey", datatypes.int_notnull)
    stats = runETL(f, datagenerator(rangeadd),
                   db_host=pg_creds.PG_HOST, db_port=pg_creds.PG_PORT,
                   db_name=pg_creds.PG_DBNAME, db_user=pg_creds.PG_USER,
                   db_pass=pg_creds.PG_PASS)
    return stats


def cleanup():
    dsn = ("dbname={dbname} user={username} password={password} host={host} port={port}"
           .format(dbname=pg_creds.PG_DBNAME, username=pg_creds.PG_USER,
                   password=pg_creds.PG_PASS, host=pg_creds.PG_HOST, port=pg_creds.PG_PORT))
    dbcon = psycopg2.connect(dsn)
    with dbcon:
        with dbcon.cursor() as cursor:
            cursor.execute("DROP SCHEMA IF EXISTS " + SCHEMA + " CASCADE")
    dbcon.close()
    print("Cleaned up")


def get_partitions(table):
    dsn = ("dbname={dbname} user={username} password={password} host={host} port={port}"
           .format(dbname=pg_creds.PG_DBNAME, username=pg_creds.PG_USER,
                   password=pg_creds.PG_PASS, host=pg_creds.PG_HOST, port=pg_creds.PG_PORT))
    dbcon = psycopg2.connect(dsn)
    partitions = get_table_partitions(dbcon, SCHEMA, table)
    dbcon.close()
    return partitions


class TestHistoricLoad(unittest.TestCase):

    def test_partitioning(self):
        etlrun(0)
        expected = []
        expected_hist = []
        for y in range(2010, 2012):
            expected_hist.append(TABLE + "_historic_" + str(y))
            for m in range(1, 13):
                expected.append(TABLE + "_" + str(y) + "_" + str(m).rjust(2, '0'))

        test = get_partitions(TABLE)
        self.assertListEqual(test, expected)

        testhist = get_partitions(TABLE + "_historic")
        self.assertListEqual(testhist, expected_hist)

        print("----- NEXT RUN -----")

        # Run next test 12 months  later. This ensures we create and delete partitions
        etlrun(12)

        expected = []
        expected_hist = []
        for y in range(2011, 2013):
            expected_hist.append(TABLE + "_historic_" + str(y))
            for m in range(1, 13):
                expected.append(TABLE + "_historic_" + str(y) + "_" + str(m).rjust(2, '0'))

    def setUp(self):
        cleanup()

    def tearDown(self):
        pass  # cleanup()


if __name__ == '__main__':
    unittest.main()



from multiprocessing import cpu_count


class Configuration():
    def __init__(self):
        self.__print_progress_line = 100000
        self.__simulate = False
        self.__debug = False
        self.__batch_size = 1000
        self.__parallel_processes = cpu_count()
        self.__batch_logging_table = True


    @property
    def print_progress_line(self):
        """Print simpleETL progress for every n lines"""
        return self.__print_progress_line

    @print_progress_line.setter
    def print_progress_line(self, value):
        if value < 1:
            raise ValueError("Print progress line must be at least 1")
        self.__print_progress_line = value

    @property
    def batch_size(self):
        """Batch size of rows distributes to each worker"""
        return self.__batch_size

    @batch_size.setter
    def batch_size(self, value):
        if value < 1:
            raise ValueError("Batch size must be at least 1")
        self.__batch_size = value

    @property
    def simulate(self):
        """Enable simulation of simpleETL. Will not commit data to fact table but dimensions can be affected"""
        return self.__simulate

    @simulate.setter
    def simulate(self, value):
        if value not in (True, False):
            raise ValueError("CONFIG.simulate value must be True og False")
        self.__simulate = value

    @property
    def debug(self):
        """Enables debug print and will not drop temporary tables"""
        return self.__debug

    @debug.setter
    def debug(self, value):
        if value not in (True, False):
            raise ValueError("CONFIG.debug value must be True og False")
        self.__debug = value

    def parallel(self, value):
        raise ValueError("CONFIG.parallel is obsolete and should not be used anymore")

    @property
    def parallel_processes(self):
        """If running in parallel, determine how many parallel process workers to spawn. Defaults to cpu_count()"""
        return self.__parallel_processes

    @parallel_processes.setter
    def parallel_processes(self, value):
        if value < 1 or value > 100:
            raise ValueError("Number of parallel processes must be in range 1-100")
        self.__parallel_processes = value

    @property
    def batch_logging_table(self):
        """Determines if simpleetl will log each batch load to a simpleetl.batchload table in the database"""
        return self.__batch_logging_table

    @batch_logging_table.setter
    def batch_logging_table(self, value):
        if value not in (True, False):
            raise ValueError("CONFIG.batch_logging_table value must be True og False")
        print("setting batch lot table to " + str(value))
        self.__batch_logging_table = value



import unittest
from simpleetl import FactTable, runETL, datatypes, CONFIG
import pg_creds
import psycopg2

SCHEMA = "testschema"
TABLE = "test_historic_load"


def datagenerator(run_no):
    for n in range(100):
        changestr = str(n * 1000 * run_no)
        yield {"lookupid": n, "teststr": changestr}


def etlrun(run_no):
    f = FactTable(schema=SCHEMA, table=TABLE, lookupatts=["lookupid"], ignore_atts=["teststr"],
                  store_history=True)
    f.handle_deleted_rows(method="wipe")
    f.add_column_mapping("lookupid", datatypes.bigint)

    f.add_column_mapping("teststr", datatypes.text)
    stats = runETL(f, datagenerator(run_no),
                   db_host=pg_creds.PG_HOST, db_port=pg_creds.PG_PORT,
                   db_name=pg_creds.PG_DBNAME, db_user=pg_creds.PG_USER,
                   db_pass=pg_creds.PG_PASS)
    return stats


def cleanup():
    dsn = ("dbname={dbname} user={username} password={password} host={host} port={port}"
           .format(dbname=pg_creds.PG_DBNAME, username=pg_creds.PG_USER,
                   password=pg_creds.PG_PASS, host=pg_creds.PG_HOST, port=pg_creds.PG_PORT))
    dbcon = psycopg2.connect(dsn)
    with dbcon:
        with dbcon.cursor() as cursor:
            cursor.execute("DROP SCHEMA IF EXISTS " + SCHEMA + " CASCADE")
    dbcon.close()
    print("Cleaned up")


class TestHistoricLoad(unittest.TestCase):

    def test_historicload(self):
        stats1_output = {'facts_inserted': 100, 'facts_updated': 0, 'facts_deleted': 0, 'historic_inserted': 100,
                         'historic_updated': 0, 'historic_deleted': 0,
                         'rows_read': 100, 'rows_parsed': 100}
        stats2_output = {'facts_inserted': 0, 'facts_updated': 0, 'facts_deleted': 0, 'historic_inserted': 0,
                         'historic_updated': 0, 'historic_deleted': 0,
                         'rows_read': 100, 'rows_parsed': 100}


        stats1 = etlrun(1)
        stats1_parsed = {key: stats1[key] for key in stats1_output.keys()}
        self.assertDictEqual(stats1_output, stats1_parsed)

        stats2 = etlrun(2)
        stats2_parsed = {key: stats2[key] for key in stats2_output.keys()}
        self.assertDictEqual(stats2_output, stats2_parsed)

    def setUp(self):
        cleanup()

    def tearDown(self):
        cleanup()


if __name__ == '__main__':
    unittest.main()



import queue
import time
import traceback
from multiprocessing import Process, Lock, Queue, Value

import psycopg2
import pygrametl
from pygrametl import ConnectionWrapper
from pygrametl.tables import BulkFactTable, _quote

from simpleetl import CONFIG
from simpleetl import LOG
from simpleetl._functions._db_tables import generate_bulkloader, getdbfriendlystr


class ETLTransform():
    def __init__(self, facttable, datafeeder, batchlog, data_timestamp, db_dsn, prefunc, processfunc):
        self.__facttable = facttable
        self.__datafeeder = datafeeder
        self.__batchlog = batchlog
        self.__data_timestamp = data_timestamp
        self.__db_dsn = db_dsn
        self.__prefunc = prefunc
        self.__processfunc = processfunc

        self.__procs_started = 0
        self.__procs_done = Value("i", 0)

        self.__lock = Lock()

        # Queue for distributing work loads to background workers
        self.__mp_data_queue = Queue(CONFIG.parallel_processes)
        self.__mp_data_queue.cancel_join_thread()  # Makes sure queue does not make main thread hang if main thread dies

        # Queue enabling background workers to communicate back used partitions to main process.
        self.__mp_partition_queue = None
        if self.__facttable.partitioning_enabled:
            self.__mp_partition_queue = Queue()
            self.__mp_partition_queue.cancel_join_thread()  # Makes sure queue does not make main thread hang if main thread dies

        # Background worker attributes
        self.__bg_worker_id = None
        self.__bg_pygramconn = None
        self.__bg_db_table = None
        self.__bg_partitions_used = None
        self.__bg_partitions_used_hist = None

    def __feeder_handler(self):

        try:
            yield from self.__datafeeder
        except Exception as err:

            tb = traceback.print_exc()
            errmsg = "ETL data feeder exception"

            LOG.error(errmsg, tb)
            self.set_shutdown()

    def __putwait(self, data, timeout=None):
        """
        :param data: Data to put
        :param timeout: Timeout in seconds
        :return:
        """
        tstart = time.time()
        while not self.is_shutdown():

            if timeout and time.time() - tstart > timeout:
                return False
            try:
                self.__mp_data_queue.put(data, timeout=1)
                return True
            except queue.Full:
                continue
        return False

    def start_etl(self):
        # Start up background dimension handlers.
        for dim in self.__facttable.dim_mappings:
            dim.dimension.connect_dim(self.__db_dsn)

        procs = []

        # Start up n background workers.
        LOG.debug("Starting up {n} parallel workers".format(n=CONFIG.parallel_processes))
        for n in range(CONFIG.parallel_processes):
            p = Process(target=self.__etl_bgworker, args=(str(n),))
            p.daemon = True
            p.start()
            procs.append(p)
            self.__procs_started += 1

        nrows = 0
        databatch = []
        t = time.time()
        n = 0
        for n, datarow_raw in enumerate(self.__feeder_handler(), 1):
            datarow = datarow_raw.copy()  # Create a shallow copy to prevent users changing data after being inserted

            if n % CONFIG.print_progress_line == 0:
                LOG.debug("Status " + str(n) + " - " + str(round(time.time() - t)))
            # If we have a prefunc, run it and let output determine if row is valid
            if self.__prefunc is not None:
                op = self.__prefunc(datarow)
                if op == False:
                    continue  # Data shall not be loaded
                elif op != True:
                    LOG.error("Prefunc must return True or False. Got: " + str(op))
                    self.set_shutdown()
                    break

            nrows += 1
            databatch.append(datarow)

            # Data is added in batches of 10000 rows. Communication is expensive.
            if len(databatch) > CONFIG.batch_size:
                if self.is_shutdown():
                    break
                did_insert = False
                while not did_insert:
                    did_insert = self.__putwait(databatch, 1)
                databatch = []
                if self.is_shutdown():
                    break
        self.__batchlog.set_rows_read(n)
        if len(databatch) > 0 and not self.is_shutdown():
            # Add any remaining data to a batch.
            self.__putwait(databatch)
        if not self.is_shutdown():
            tsstr = time.strftime("%H:%M:%S", time.gmtime(time.time() - t))
            LOG.info("All data, " + str(n) + " rows, loaded into workers in " + tsstr)

        # Keep waiting until all batches have been processed
        ends_put = 0
        LOG.debug("Waiting for processes to finish "+str(self.is_shutdown()))
        tstart = time.time()
        while self.__procs_done.value < self.__procs_started and not self.is_shutdown():
            # LOG.debug(str(("ends_put",ends_put,self.__procs_done,self.__procs_started)))
            while ends_put < self.__procs_started:
                success = self.__putwait("END", 300)
                # LOG.debug("putwait test "+str(success))
                if not success:
                    LOG.error(
                        "Background worker not finishing. Queue not getting emptied. " + str(self.__procs_done.value)
                        + " of " + str(self.__procs_started) + " has finished", None)
                    if self.set_shutdown():
                        break
                else:
                    ends_put += 1
            time.sleep(1)
            if time.time() - tstart > 300:
                LOG.error("Background worker not finishing. " + str(self.__procs_done.value)
                          + " of " + str(self.__procs_started) + " has finished", None)
                self.set_shutdown()

        LOG.debug("Cleaning up")
        self.__cleanup_queues()
        if (not self.is_shutdown() and (
                self.__facttable.partitioning_enabled or self.__facttable.partitioning_enabled_hist)):
            while True:
                try:
                    pset, pset_hist = self.__mp_partition_queue.get(timeout=1)
                    for p in pset:
                        self.__facttable.add_used_partition(p)
                    for p in pset_hist:
                        self.__facttable.add_used_partition_hist(p)
                except queue.Empty:
                    break

        if self.is_shutdown():
            LOG.error("Aborting ETL due to error")
            for dim in self.__facttable.dim_mappings:
                if CONFIG.debug: LOG.error("Aboring dimension: " + dim.dimension._get_full_table())
                dim.dimension.abort()
            if CONFIG.debug: LOG.error("Returning from Transform")
            return False

        for dim in self.__facttable.dim_mappings:
            dim.dimension.close()

        tsstr = time.strftime("%H:%M:%S", time.gmtime(time.time() - t))
        LOG.info("All workers done in " + tsstr)
        self.__batchlog.set_rows_parsed(nrows)
        return True

    def __cleanup_queues(self):

        LOG.debug("Emptying and closing data queue")

        while True:
            try:
                self.__mp_data_queue.get_nowait()
            except queue.Empty:
                break

        LOG.debug("Cleanup done")

    def __etl_bgworker(self, worker_id):

        self.__bg_worker_id = worker_id

        self.__bg_partitions_used = set()
        self.__bg_partitions_used_hist = set()
        try:
            while True:

                if self.is_shutdown():
                    break
                try:
                    data = self.__mp_data_queue.get(timeout=1)

                    if data == "END":

                        if self.__bg_pygramconn:
                            self.__bg_pygramconn.commit()
                            self.__bg_pygramconn.close()
                            self.__bg_pygramconn = None
                        break
                    try:
                        status = self.__transform(data)
                        # Status true means everything is okay
                        if isinstance(status, str):
                            tb = traceback.format_exc()
                            LOG.error(status, tb)
                            self.set_shutdown()
                            break
                        elif status == False:
                            break
                    except Exception as err:
                        tb = traceback.format_exc()
                        LOG.error("Background ETL worker failed", tb)
                        self.set_shutdown()

                        break
                except queue.Empty:
                    continue
            if self.__facttable.partitioning_enabled or self.__facttable.partitioning_enabled_hist:
                # Return used partitions to main process, if partitioning is enablefs.
                self.__mp_partition_queue.put((self.__bg_partitions_used, self.__bg_partitions_used_hist))

        except Exception:
            tb = traceback.format_exc()

            LOG.error("Background ETL worker failed", tb)
            self.set_shutdown()
        finally:
            if self.__bg_pygramconn:
                try:
                    self.__bg_pygramconn.rollback()
                    self.__bg_pygramconn.close()
                except Exception:
                    pass
            with self.__lock:
                self.__procs_done.value += 1

    def __transform(self, rows):
        if self.__bg_db_table is None:
            pygrametl.tables.definequote('"')  # Tell pygrametl to quote identifiers.
            name = "SimpleETL BG worker {s}.{t} ({id})".format(s=self.__facttable.schema,
                                                               t=self.__facttable.table,
                                                               id=self.__bg_worker_id)
            pgconn = psycopg2.connect(self.__db_dsn, application_name=name)
            self.__bg_pygramconn = ConnectionWrapper(pgconn)
            keycols = self.__facttable._get_cols(only_lookupatts=True)
            measurecols = self.__facttable._get_cols(include_lookupatts=False, include_ignorecols=True,
                                                     include_changeatts=True)

            self.__bg_db_table = BulkFactTable(name=_quote("tmp_data") + "." +
                                                    _quote(self.__facttable.tmptable),
                                               keyrefs=keycols,
                                               measures=measurecols,
                                               bulkloader=generate_bulkloader(self.__bg_pygramconn),
                                               fieldsep='\t',
                                               rowsep='\n',
                                               nullsubst="\\N",
                                               tempdest=None,
                                               bulksize=100000,
                                               usefilename=False,
                                               dependson=(),
                                               strconverter=getdbfriendlystr)
        n = 0
        """ Variables for accessing facttable object. Variable access is faster than object access"""
        dim_mappings = self.__facttable.dim_mappings
        col_mappings = self.__facttable.simple_mappings
        lookupatts_mappings = self.__facttable._get_cols(only_lookupatts=True)
        marked_deleted = self.__facttable.deleted_rows_method == "mark"
        track_last_updated = self.__facttable.track_last_updated
        track_created = self.__facttable.track_created
        has_processfunc = self.__processfunc is not None
        partitioning_enabled = self.__facttable.partitioning_enabled
        partitioning_enabled_hist = self.__facttable.partitioning_enabled_hist
        factinsert = self.__bg_db_table.insert

        for srcrow in rows:
            n += 1

            destrow = {}
            # If removed rows should be marked, we here add that this row is not deleted.
            # This is necessary, as _deleted is treated as a regular data column and a value is expected.
            if marked_deleted:
                srcrow["_deleted"] = None
            if track_last_updated:
                srcrow["_updated"] = self.__data_timestamp
            if track_created:
                srcrow["_created"] = self.__data_timestamp
            # If any user-defined functionality is provided, execute this first.
            if has_processfunc:
                procoutp = self.__processfunc(srcrow)
                if procoutp is False:
                    continue

            # Looking up values in lookup-only dimensions.
            # This is not inserted but could be used by other dimensions
            for dim in dim_mappings:
                dimkey = dim.dimension.ensure(srcrow, dim.namemapping)
                if dimkey is None:
                    return "No dimension key found for " + str(dim)
                if isinstance(dimkey, tuple):
                    return "Error when handling dimension " + str(dim) + ": " + str(dimkey)
                if dim.insert:
                    destrow[dim.dstcol] = dimkey
                if dim.inject:
                    srcrow[dim.dstcol] = dimkey
            # One-to-one mapping of data from source data to fact column
            for simpcol in col_mappings:
                try:
                    if srcrow[simpcol.srccol] is None:
                        destrow[simpcol.dstcol] = None
                    else:
                        # Add data to destination dictionary, parse by type.
                        destrow[simpcol.dstcol] = simpcol.datatype.parse(srcrow[simpcol.srccol])
                except KeyError as err:
                    return ("Cannot define attribute " + simpcol.dstcol +
                            " - Column " + simpcol.srccol + " missing in source data: " + str(err))
                except Exception as err:
                    return ("Something went bad with " + simpcol.srccol + ". Content: " + str(
                        srcrow[simpcol.srccol]) + ". Error:: " + str(err))
            for lookupcol in lookupatts_mappings:
                if destrow[lookupcol] is None:
                    return ("Column " + lookupcol) + " is part of lookup attributes but value is NULL!"
            # If partition is enabled, add current partition to set of partitions
            if partitioning_enabled:
                p = self.__facttable.get_partition(destrow)
                self.__bg_partitions_used.add(p)
            if partitioning_enabled_hist:
                p = self.__facttable.get_partition_hist(destrow)
                self.__bg_partitions_used_hist.add(p)

            factinsert(destrow)
        return None



import os
import queue
import sys
import traceback
from collections import namedtuple
from functools import lru_cache
from multiprocessing import Process, Queue, Lock

import psycopg2
from pygrametl import getvalue, ConnectionWrapper
from pygrametl.tables import CachedDimension

from simpleetl import LOG, CONFIG
from simpleetl._functions import _datatypes
from simpleetl._functions._db_quoting import quote, quotelist
from simpleetl._functions._db_schema_handler import addcol
from simpleetl._functions._dbfunctions import table_exists, has_index, get_lock_transaction
from simpleetl._processing.ETLProcess import ETLProcess

# Data structure for mapping field between raw data and
# table column, including data type
Column = namedtuple("Column", ["srccol", "dstcol", "datatype"])

# Data structure for mapping a source data field to a dimension
# dimension is dimension object, dstcol is destination column in fact table,
# namemapping is a mapping {dstname: sourcename} where dstname is a column name used by the
# dimension object and sourcename is the name from the data source.
# Inject defines if the found key value from the dimension will be injected into the source data, e.g., to be used
# by other dimensions.
DimColumn = namedtuple("DimColumn", ["dimension", "dstcol", "namemapping", "inject", "insert", "fkey", "allow_null"])

# Dictionaries are ordered by creation by default from python 3.6
dict_ordered = (sys.version_info.major == 3 and sys.version_info.minor >= 6)


def _quote(schema, table): return '"' + schema + '"."' + table + '"'


class Dimension:
    """Dimension object
    """

    def __init__(self, schema, table, key, rowexpander=None,
                 cachesize=100000, integerkey=False):
        """
        :param schema: Schema for dimension.
        :param table: Table for dimension.
        :param key: Primary key of table.
        :param rowexpander: Rowexpander functionality for expanding data to. Must be dict
        :param cachesize: Size of cache.
        :param integerkey: Use integer instead of (default) bigint,
        """

        self.__schema = schema
        self.__table = table
        self.__key = key
        self.__rowexpander = rowexpander
        self.__cachesize = cachesize
        self.__mainpid = os.getpid()

        if integerkey == True:
            self.__keytype = _datatypes.int_notnull

        else:
            self.__keytype = _datatypes.bigint_notnull

        self.__atts = []
        self.__lookupatts = []
        self.__proc = None

        self.__dim = None

        self.__qin = Queue(100)
        self.__qin.cancel_join_thread()  # Makes sure queue does not make main thread hang if main thread dies
        self.__qout = Queue(100)
        self.__qout.cancel_join_thread()  # Makes sure queue does not make main thread hang if main thread dies
        self.__lock = Lock()

        self.__hits = 0
        self.__bghits = 0

        self.__pygramconn = None
        self.__pgconn = None

    def add_lookupatt(self, name, dtype, default_value=chr(0)):
        """
        Add lookup attribute for dimension
        :param name: Name of destination column
        :param dtype: Datatype
        :param default_value: Allow substitution with this value when missing.
            If NULL CHR (chr(0)), do not allow missing values
        :return:
        """
        if dtype.allow_null:
            dtype = dtype.copy(False)  # Ensure we disallow NULL
        if name == self.__key:
            # If we use a smart key, i.e. user-defined primary key, update the datatype according.
            self.__keytype = dtype
        self.__lookupatts.append((name, dtype, default_value))

    def get_lookupatts(self):
        return self.__lookupatts.copy()

    def add_att(self, name, dtype, default_value=chr(0)):
        self.__atts.append((name, dtype, default_value))

    def get_lookuptype(self, name):
        for lname, ltype, defval in self.__lookupatts:
            if lname == name:
                return ltype
        raise ValueError("Dimension.get_lookuptype (" + self.__schema + "." + self.__table +
                         "): Did not find name " + name + " in " + str(self.__lookupatts))

    def connect_dim(self, db_dsn, errorqueue=None):
        # If we only have one lookup att and it is the same as the key, we have a
        # smart key. Thus we do not need to need to wait for return values from the
        # dimension but we can compute this ourselves.
        """
        :param db_dsn: Database connectionstring
        :param errorqueue: Message queue to tell main process if any errors happens.
        :return:
        """
        if len(self.__lookupatts) == 0:
            raise AttributeError("No lookup atts added to dimension " + self._get_full_table())

        if errorqueue is None:
            class custq:
                def __init__(self):
                    pass

                def put(self, data):
                    print(data)

            errorqueue = custq()
        self.__db_dsn = db_dsn
        self.__errorqueue = errorqueue

        if self.__proc is None:
            self.__proc = Process(target=self.__bgworker)
            self.__proc.daemon = True
            self.__proc.start()

    def _get_schema(self):
        return self.__schema

    def _get_table(self):
        return self.__table

    def _get_full_table(self):
        return _quote(self.__schema, self.__table)

    def _get_key(self):
        return self.__key

    def _get_keytype(self):

        return self.__keytype

    def create_schema(self, pgconn):
        if len(self.__lookupatts) == 0:
            raise AttributeError("No lookup atts added to dimension " + self._get_full_table())
        # Backward compatibility, when dsn was given instead of pgconn
        close_pgconn = False
        if isinstance(pgconn, str):
            pgconn = psycopg2.connect(pgconn, application_name="Dim ensure: " + self._get_full_table())
            close_pgconn = True

        if not table_exists(pgconn, self.__schema, self.__table):
            tabcreate = """CREATE SCHEMA if not exists {sch};
                CREATE TABLE if not exists {sch}.{tab}
                (  {key} {keytype},
                   PRIMARY KEY ({key}))
            """.format(sch=quote(self.__schema, pgconn), tab=quote(self.__table, pgconn), key=quote(self.__key, pgconn),
                       keytype=self.__keytype.sqltype())
            with pgconn.cursor() as cursor:
                cursor.execute(tabcreate)

        uniqueidx = []

        for col, datatype, _ in self.__lookupatts:
            uniqueidx.append(col)
            # self.__addcol(pgconn, col, datatype)
            addcol(self.__schema, self.__table, pgconn, col, datatype)
        for col, datatype, _ in self.__atts:
            # self.__addcol(pgconn, col, datatype, True)
            addcol(self.__schema, self.__table, pgconn, col, datatype)
        lookupidx = uniqueidx.copy()
        lookupidx.append(self.__key)

        hasidx = has_index(pgconn, self.__schema, self.__table, lookupidx)
        if not hasidx:
            idx = """CREATE unique INDEX ON {schema}.{table}
                                        ({keycols});
                                        """.format(schema=quote(self.__schema, pgconn),
                                                   table=quote(self.__table, pgconn),
                                                   keycols=", ".join(quotelist(lookupidx, pgconn)))

            with pgconn.cursor() as cursor:
                cursor.execute(idx)
        from simpleetl._functions._dbfunctions import has_unique, create_unique
        if not has_unique(pgconn, self.__schema, self.__table, uniqueidx):
            create_unique(pgconn, self.__schema, self.__table, uniqueidx)
        pgconn.commit()
        if close_pgconn:
            pgconn.close()

    def close(self):
        if self.__proc is not None:
            self.__qin.put("STOP")
            h, bgh = self.__qout.get(timeout=5)
            self.__proc.join()
        self.__proc = None

    def abort(self):
        if self.__proc is not None:
            self.__qin.put("ABORT")
            self.__proc.join()

            self.__hits = 0
            self.__bghits = 0
            self.__proc = None

    """ ---------- The following methods are methods run by parallel background worker ETL processes ---------- """

    def ensure(self, srcdata, namemapping={}):
        """
        Ensure a dimension attribute exists and return key.
        :param srcdata:
        :param namemapping:
        :return:
        """

        # Start up background worker, in case not started
        if not self.__proc:
            LOG.error(
                "Ensure called on dimension, but dimension has not been connected yet: " + self.__schema + "." + self.__table)

        # Construct a lookup row for the dimension, given the lookup attributes.
        dimrow = {}

        for colname, datatype, default_value in self.__lookupatts:
            try:
                k = getvalue(srcdata, colname, namemapping)
            except KeyError as err:
                errstr = (
                        "Dimension " + self._get_full_table() + " requires lookup attribute " + colname + ". This was not found in source data: " + str(
                    srcdata.keys()))
                LOG.error(errstr)
                # raise KeyError("dimension line ~124: " + self.get_full_table() + " ensure error: " + str(err))
                return ((errstr, err))

            if k is None:
                if default_value is not chr(0):
                    k = default_value
                else:
                    errstr = (
                            "Lookup-attribute is None in dimension for lookup-attribute" + colname + " in " + self._get_full_table() + ". Value '" + str(
                        k) + "' parsed to 'None'")
                    LOG.error(errstr)
                    return ((errstr, None))
            try:
                outval = datatype.parse(k)
            except AttributeError as err:
                errstr = (
                        "Lookup-attribute is wrong type in dimension for lookup-attribute" + colname + " in " + self._get_full_table() + ". Value '" + str(
                    k) + "' parsed to 'None'. Error: " + str(err))
                LOG.error(errstr)
                return ((errstr, err))

            if outval is None:
                errstr = (
                        "Could not parse dimensioon lookup-attribute" + colname + " in " + self._get_full_table() + ". Value '" + str(
                    k) + "' parsed to '" + str(outval) + "'")
                LOG.error(errstr)
                return ((errstr, None))

            dimrow[colname] = outval

        self.__hits += 1
        for colname, datatype, default_value in self.__atts:
            try:
                k = getvalue(srcdata, colname, namemapping)
            except KeyError as err:
                if default_value is not chr(0):
                    k = default_value
                else:
                    continue

            if k is None:
                if default_value is not chr(0):
                    k = default_value
                else:
                    continue
            outval = datatype.parse(k)
            if outval is None:
                errstr = (
                        "Could not parse dimension attribute " + colname + " in " + self._get_full_table() + ". Value '" + str(
                    k) + "' parsed to '" + str(outval) + "'")
                LOG.error(errstr)
                return (errstr, None)
            dimrow[colname] = outval

        # Create a frozen set, which can be cached, from the input dictionary
        dimrow = frozenset(dimrow.items())
        return self.__decoupled_return_ensure(dimrow)

    @lru_cache(100000)
    def __decoupled_return_ensure(self, data):
        """
        :param datamsg:
        :param row: Input data, can be either a message, "STOP" or "ABORT", or a frozenset of data to look up.
        :return: Dimension key returned from background worker
        """
        with self.__lock:
            self.__qin.put(data)
            return self.__qout.get()

    """ ---------- The following methods are background worker methods ---------- """

    def __bgworker(self):
        """
        Background worker, handling dimension database action.
        """
        self.__connect_bgworker()
        abort = False

        while True:
            try:
                # Get entry from queue. Wait up to three seconds, before testing if something is wrong
                data = self.__qin.get(timeout=3)
            except queue.Empty:
                try:
                    os.kill(self.__mainpid, 0)  # Test if main process is alive. Throws PermissionError if not
                    if ETLProcess.is_shutdown():
                        data = "ABORT"
                    else:
                        continue
                except PermissionError:
                    # Main process is dead
                    data = "ABORT"

            if data == "STOP":
                break
            elif data == "ABORT":
                abort = True
                break
            else:
                self.__hits += 1
                v = self.__do_map(data)
                self.__qout.put(v)

        if not CONFIG.simulate and not abort:
            self.__pygramconn.commit()
        else:
            self.__pygramconn.rollback()
        self.__pygramconn.close()

        self.__qout.put((self.__hits, self.__bghits))

    def __connect_bgworker(self):
        self.__pgconn = psycopg2.connect(self.__db_dsn, application_name="SimpleETL Dimension {s}.{t}".format(
            s=self.__schema, t=self.__table))
        self.__pygramconn = ConnectionWrapper(self.__pgconn)

        atts = [a[0] for a in self.__lookupatts + self.__atts]
        if self.__key in atts:
            atts.remove(self.__key)

        def _getnextid(ignoredrow, ignoredmapping):
            """
            Function for always fetching greatest ID from DB table. This enables parallel ETL as multiple dimension
            objects on the same table can run in parallel.
            :param ignoredrow:
            :param ignoredmapping:
            :return:
            """
            with self.__pygramconn.cursor() as cursor:
                cursor.execute("SELECT coalesce(MAX({id}),0)+1 as id FROM {t}".format(id=self.__key,
                                                                                      t=self._get_full_table()))
                row = cursor.fetchone()
            return row[0]

        self.__dim = CachedDimension(name=self._get_full_table(),
                                     key=self.__key,
                                     lookupatts=[a[0] for a in self.__lookupatts],
                                     attributes=atts,
                                     size=self.__cachesize,
                                     usefetchfirst=True,
                                     prefill=True,
                                     idfinder=_getnextid,
                                     cacheoninsert=False,  # Very important, this enforces to look in the DB for data.
                                     rowexpander=self.__rowexpander,
                                     targetconnection=self.__pygramconn)
        self.__pygramconn.commit()

    @lru_cache(10000)
    def __do_map(self, inputrow):
        try:
            row = dict(inputrow)
            keyval = self.__dim.lookup(row)
            if keyval is None:
                lockkey = self.__schema + "." + self.__table
                self.__bghits += 1
                # Require a transaction lock on the dimension table. This ensures other ETL processes can use the same
                # dimension in parallel, but inserted entries will immediately be available to other ETL processes.
                # Lock is released upron commit or rollback.
                get_lock_transaction(self.__pgconn, lockkey)

                keyval = self.__dim.ensure(row)
                if not CONFIG.simulate:
                    # Other seperate ETL jobs can utilized the same dimension. Therefore we need to commit changes immediately to the DB
                    self.__pygramconn.commit()
                else:
                    self.__pygramconn.rollback()

        except:
            err = sys.exc_info()[0]
            keyval = ("Error", str(err) + " - " + str(traceback.print_exc()))

        return keyval



import datetime
import json
import uuid
import xml.etree.ElementTree as ET
from decimal import Decimal


def is_timestamp(x):
    if isinstance(x, datetime.datetime) or x is None:
        return x

    raise ValueError("is_timestamp should receive datetime.datetime, input type is " + str(type(x)))


def is_timestamptz(x):
    dt = is_timestamp(x)
    if dt is None or (dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None):
        return dt
    raise ValueError("Time stamp has not time zone information: " + str(x))


def is_timekey(x):
    if isinstance(x, int):
        return x
    raise ValueError("timekey should receive integer representation, input type is " + str(type(x)))

    return None


def is_datekey(x):
    if isinstance(x, int):
        return x

    raise ValueError("datekey should receive pre-processed integer datekey, input type is " + str(type(x)))


def generate_is_varchar(length):
    def is_varchar(x):
        x = is_text(x)
        if len(x) > length:
            raise ValueError("varchar(" + str(length) + ") overflow: " + x)
        return x

    return is_varchar


def is_text(x, postfunc=None):
    #x = x.replace("\t", "\\t").replace("\n", "\\n").replace("\r", "\\r").strip()
    x = x.strip()
    if x is not None and postfunc:
        x = postfunc(x)
    return x


def is_textlower(x):
    return is_text(x, str.lower)


def is_textupper(x):
    return is_text(x, str.upper)


def is_bigint(x):
    if not isinstance(x, int):
        x = str(x).strip()
        try:
            x = int(x)
        except ValueError:
            return None
    if -9223372036854775808 <= x <= 9223372036854775807:
        return x
    else:
        raise ValueError("Bigint out of range", x)


def is_int(x):
    x = is_bigint(x)
    if x is None:
        return None
    if -2147483648 <= x <= 2147483647:  # We only want 32 bit signed integers
        return x
    else:
        raise ValueError("Integer out of range", x)


def is_boolean(x):
    if x == False or x == True:
        return x
    else:
        raise ValueError("Invalid boolean", x)


def is_uuid(x):
    try:
        return str(uuid.UUID(x))
    except ValueError as err:
        raise ValueError("Invalid UUID", x)


def is_intarray(x):
    if not isinstance(x, list):
        raise ValueError("Intarray must be list, got", type(x))
    data = str([str(int(n)) for n in x])

    return data.replace('[', '{').replace(']', '}').replace('\'', '\"')


################################################################################
# spatial
################################################################################
def is_latitude(x):
    try:
        x = float(x)
        if -90 <= x <= 90:
            return x
        else:
            raise ("Out of Earth latitude", x)
    except ValueError:
        raise ValueError("Not a float", x)


def is_longitude(x):
    try:
        x = float(x)
        if -180 <= x <= 180:
            return x
        else:
            raise ("Out of Earth logitude", x)
    except ValueError:
        raise ValueError("Not a float", x)


def is_altitude(x):
    try:
        x = float(x)
        if -1000 <= x <= 10000:
            return x
        else:
            raise ("Out of Earth altitude", x)
    except ValueError:
        raise ValueError("Not a float", x)


def is_degree(x):
    try:
        x = int(x)
        if 0 <= x < 360:
            return x
        else:
            raise ("Degree not in 0-360 range", x)
    except ValueError:
        raise ValueError("Not an integer", x)


################################################################################
# XML Schema Inspired Primitive Data Types
################################################################################

def is_non_negative_int(x):
    x = is_bigint(x)
    if x is None:
        return None
    if 0 <= x <= 2147483647:
        return x
    else:
        raise ValueError("Non-negative integer value out of range", x)


def is_non_positive_int(x):
    x = is_bigint(x)
    if x is None:
        return None
    if -2147483648 <= x <= 0:
        return x
    else:
        raise ValueError("Non-negative integer value out of range", x)


def is_positive_int(x):
    x = is_bigint(x)
    if x is None:
        return None
    if 1 <= x <= 2147483647:
        return x
    else:
        raise ValueError("Non-negative integer value out of range", x)


def is_smallint(x):
    x = is_int(x)
    if x is None:
        return None
    if -32768 <= x <= 32767:  # We only want 16bit signed integers
        return x
    else:
        raise ValueError("Smallint out of range", x)


def generate_is_numeric(precission, scale):
    digs = 10 ** (precission - scale)

    def is_numeric(x):
        if isinstance(x, str):
            x = x.strip()
        else:
            x = str(x)
        if x:
            dec = Decimal(x.replace(",", "."))
            if -digs >= dec or dec >= digs:
                raise ValueError("Numeric(" + str(precission) + "," + str(scale) + ") overflow: " + str(dec))
            return dec
        return None

    return is_numeric


def injson(x):
    try:
        if isinstance(x, str):
            x = json.loads(x)
        return json.dumps(x).replace('\\', '\\\\')

    except ValueError as err:
        print("Ill formated json", err, x)
    return None


def inxml(x):
    try:
        ET.fromstring(x)
        return x

    except ET.ParseError as err:
        print("Ill formated XML", err, x)
    return None



"""
   The package's modules are:
   - FactTable for defining a fact table
   - Dimension for defining a dimension, which can be attached a FactTable object
   - Datatype for letting users define their own datatypes
   - LOG for using the build-in log functionality from simpleetl
   - CONFIG for setting global configuration parameters
   - runETL for initiating an ETL batch run
   - datatypes giving access to pre-defined datatypes.
   - datatypefuncs giving access to pre-defined functions for defining new Datatype objects
"""

# Copyright (c) 2020, FlexDanmark
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__all__ = (
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    'FactTable',
    'Dimension',
    'Datatype',
    'LOG',
    'CONFIG',
    'runETL',
    'datatypes',
    'datatypefuncs')

from simpleetl._modules.Configuration import Configuration as __Configuration

CONFIG = __Configuration()
from simpleetl._modules.Logger import LOGCLASS

LOG = LOGCLASS()

from simpleetl._modules.Datatype import Datatype
from simpleetl._functions import _datatypes as datatypes
from simpleetl._functions import _datatypefuncs as datatypefuncs
from simpleetl._modules.Dimension import Dimension

from simpleetl._processing.runetl import runETL

from simpleetl._modules.FactTable import FactTable

import sys

if sys.version_info[:2] >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata

metadata = importlib_metadata.metadata("simpleetl")

__title__ = metadata["name"]
__summary__ = metadata["summary"]
__uri__ = metadata["home-page"]
__version__ = metadata["version"]
__author__ = metadata["author"]
__email__ = metadata["author-email"]
__license__ = metadata["license"]



"Clone of unreleased pygrametl table."

import pygrametl
from pygrametl.tables import Dimension, FIFODict


class SlowlyChangingDimension(Dimension):
    """A class for accessing a slowly changing dimension of "type 2".

       "Type 1" updates can also be applied for a subset of the attributes.

       Caching is used. We assume that the DB doesn't change or add any
       attribute values that are cached.
       For example, a DEFAULT value in the DB or automatic type coercion can
       break this assumption.
    """

    def __init__(self, name, key, attributes, lookupatts, versionatt,
                 fromatt=None, fromfinder=None,
                 toatt=None, tofinder=None, minfrom=None, maxto=None,
                 srcdateatt=None, srcdateparser=pygrametl.ymdparser,
                 type1atts=(), cachesize=10000, prefill=False, idfinder=None,
                 usefetchfirst=False, useorderby=True, targetconnection=None):
        """Arguments:
           - name: the name of the dimension table in the DW
           - key: the name of the primary key in the DW
           - attributes: a sequence of the attribute names in the dimension
             table. Should not include the name of the primary key which is
             given in the key argument, but should include versionatt,
             fromatt, and toatt.
           - lookupatts: a sequence with a subset of the attributes that
             uniquely identify a dimension members. These attributes are thus
             used for looking up members.
           - versionatt: the name of the attribute holding the version number
           - fromatt: the name of the attribute telling from when the version
             becomes valid. Not used if None. Default: None
           - fromfinder: a function(targetconnection, row, namemapping)
             returning a value for the fromatt for a new version (the function
             is first used when it is determined that a new version must be
             added; it is not applied to determine this).
             If fromfinder is None and srcdateatt is also None,
             pygrametl.today is used as fromfinder. If fromfinder is None
             and srcdateatt is not None,
             pygrametl.datereader(srcdateatt, srcdateparser) is used.
             In other words, if no date attribute and no special
             date function are given, new versions get the date of the current
             day. If a date attribute is given (but no date function), the
             date attribute's value is converted (by means of srcdateparser)
             and a new version gets the result of this as the date it is valid
             from. Default: None
           - toatt: the name of the attribute telling until when the version
             is valid. Not used if None. Default: None
           - tofinder: a function(targetconnection, row, namemapping)
             returning a value for the toatt. If not set, fromfinder is used
             (note that if fromfinder is None, it is set to a default
             function -- see the comments about fromfinder. The possibly
             modified value is used here.) Default: None
           - minfrom: the value to use for fromatt for the 1st version of a
             member if fromatt is not already set. If None, the value is
             found in the same way as for other new versions, i.e., as
             described for fromfinder. If fromatt should take the value
             NULL for the 1st version, set minfrom to a tuple holding a single
             element which is None: (None,). Note that minto affects the 1st
             version, not any following versions. Note also that if the member
             to insert already contains a value for fromatt, minfrom is
             ignored.
             Default: None.
           - maxto: the value to use for toatt for new members. Default: None
           - srcdateatt: the name of the attribute in the source data that
             holds a date showing when a version is valid from. The data is
             converted to a datetime by applying srcdateparser on it.
             If not None, the date attribute is also used when comparing
             a potential new version to the newest version in the DB.
             If None, the date fields are not compared. Default: None
           - srcdateparser: a function that takes one argument (a date in the
             format scrdateatt has) and returns a datetime.datetime.
             If srcdateatt is None, srcdateparser is not used.
             Default: pygrametl.ymdparser (i.e., the default value is a
             function that parses a string of the form 'yyyy-MM-dd')
           - type1atts: a sequence of attributes that should have type1 updates
             applied. Default: ()
           - cachesize: the maximum size of the cache. 0 disables caching
             and values smaller than 0 allows unlimited caching
           - prefill: decides if the cache should be prefilled with the newest
             versions. Default: False.
           - idfinder: a function(row, namemapping) -> key value that assigns
             a value to the primary key attribute based on the content of the
             row and namemapping. If not given, it is assumed that the primary
             key is an integer, and the assigned key value is then the current
             maximum plus one.
           - usefetchfirst: a flag deciding if the SQL:2008 FETCH FIRST
             clause is used when prefil is True. Depending on the used DBMS
             and DB driver, this can give significant savings wrt. to time and
             memory. Not all DBMSs support this clause yet. Default: False
           - targetconnection: The ConnectionWrapper to use. If not given,
             the default target connection is used.
           - useorderby: a flag deciding if ORDER BY is used in the SQL to
             select the newest version. If True, the DBMS thus does the
             sorting. If False, all versions are fetched and the highest
             version is found in Python. For some systems, this can lead to
             significant performance improvements. Default: True
        """
        # TODO: Should scdensure just override ensure instead of being a new
        #       method?

        Dimension.__init__(self,
                           name=name,
                           key=key,
                           attributes=attributes,
                           lookupatts=lookupatts,
                           idfinder=idfinder,
                           defaultidvalue=None,
                           rowexpander=None,
                           targetconnection=targetconnection)

        if not versionatt:
            raise ValueError('A version attribute must be given')

        self.versionatt = versionatt
        self.fromatt = fromatt
        if fromfinder is not None:
            self.fromfinder = fromfinder
        elif srcdateatt is not None:  # and fromfinder is None
            self.fromfinder = pygrametl.datereader(srcdateatt, srcdateparser)
        else:  # fromfinder is None and srcdateatt is None
            self.fromfinder = pygrametl.today
        self.toatt = toatt
        if tofinder is None:
            tofinder = self.fromfinder
        self.tofinder = tofinder
        self.minfrom = minfrom
        self.maxto = maxto
        self.srcdateatt = srcdateatt
        self.srcdateparser = srcdateparser
        self.type1atts = type1atts
        self.useorderby = useorderby
        if cachesize > 0:
            self.rowcache = FIFODict(cachesize)
            self.keycache = FIFODict(cachesize)
        elif cachesize < 0:
            self.rowcache = {}
            self.keycache = {}
        # else cachesize == 0 and we do not create any caches
        self.__cachesize = cachesize
        self.__prefill = cachesize and prefill  # no prefilling if no caching

        # Check that versionatt, fromatt and toatt are also declared as
        # attributes
        for var in (versionatt, fromatt, toatt):
            if var and var not in attributes:
                raise ValueError("%s not present in attributes argument" %
                                 (var,))

        # Now extend the SQL from Dimension such that we use the versioning
        self.keylookupsql += " ORDER BY %s DESC" % (self.quote(versionatt),)

        # Now create SQL for looking up the key with a local sort
        # This gives "SELECT key, version FROM name WHERE
        # lookupval1 = %(lookupval1)s AND lookupval2 = %(lookupval2)s AND ..."
        self.keyversionlookupsql = "SELECT " + self.quote(key) + ", " + \
                                   self.quote(versionatt) + " FROM " + name + \
                                   " WHERE " + " AND ".join(["%s = %%(%s)s" % (self.quote(lv), lv)
                                                             for lv in lookupatts])

        if toatt:
            self.updatetodatesql = \
                "UPDATE %s SET %s = %%(%s)s WHERE %s = %%(%s)s" % \
                (name, self.quote(toatt), toatt, self.quote(key), key)

        if self.__prefill:
            self.__prefillcaches(usefetchfirst)

    def __prefillcaches(self, usefetchfirst):
        args = None
        if self.toatt:
            # We can use the toatt to see if rows are still current.
            # Select all attributes from the rows where maxto is set to the
            # default value (which may be NULL)
            sql = 'SELECT %s FROM %s WHERE %s %s' % \
                  (', '.join(self.quotelist(self.all)), self.name,
                   self.quote(self.toatt),
                   self.maxto is None and 'IS NULL' or '= %(maxto)s')
            if self.maxto is not None:
                args = {'maxto': self.maxto}
        else:
            # We have to find max(versionatt) for each group of lookupatts and
            # do a join to get the right rows.
            lookupattlist = ', '.join(self.lookupatts)
            newestversions = ('SELECT %s, MAX(%s) AS %s FROM %s GROUP BY %s' %
                              (self.quote(lookupattlist),
                               self.quotelist(self.versionatt),
                               self.quotelist(self.versionatt), self.name,
                               self.quotelist(lookupattlist)))
            joincond = ' AND '.join(['A.%s = B.%s' % (self.quote(att), att)
                                     for att in [l for l in self.lookupatts] +
                                     [self.versionatt]
                                     ])
            sql = 'SELECT %s FROM (%s) AS A, %s AS B WHERE %s' % \
                  (', '.join(['B.%s AS %s' % (self.quote(att), att)
                              for att in self.all]),
                   newestversions, self.name, joincond)

        # sql is a statement that fetches the newest versions from the database
        # in order to fill the caches, the FETCH FIRST clause is for a finite
        # cache, if the user set the flag that it is supported by the database.
        positions = [self.all.index(att) for att in self.lookupatts]
        if self.__cachesize > 0 and usefetchfirst:
            sql += ' FETCH FIRST %d ROWS ONLY' % self.__cachesize
        self.targetconnection.execute(sql, args)

        if self.__cachesize < 0:
            allrawrows = self.targetconnection.fetchalltuples()
        else:
            allrawrows = self.targetconnection.fetchmanytuples(
                self.__cachesize)

        for rawrow in allrawrows:
            self.rowcache[rawrow[0]] = rawrow
            t = tuple([rawrow[i] for i in positions])
            self.keycache[t] = rawrow[0]

    def lookup(self, row, namemapping={}):
        """Find the key for the newest version with the given values.

           Arguments:
           - row: a dict which must contain at least the lookup attributes
           - namemapping: an optional namemapping (see module's documentation)
        """
        if self.__prefill and (self.__cachesize < 0 or
                               len(self.keycache) < self.__cachesize):
            # Everything is cached. We don't have to look in the DB
            return self._before_lookup(row, namemapping)
        else:
            # Something is not cached so we have to use the classical lookup.
            # Note that __init__ updated self.keylookupsql to use ORDER BY ...
            if self.useorderby:
                return Dimension.lookup(self, row, namemapping)
            else:
                return self.__lookupnewestlocally(row, namemapping)

    def __lookupnewestlocally(self, row, namemapping):
        """Find the key for the newest version of the row with the given values.

           Arguments:
           - row: a dict which must contain at least the lookup attributes
           - namemapping: an optional namemapping (see module's documentation)
        """
        # Based on Dimension.lookup, but uses keyversionlookupsql and
        # finds the newest version locally (no sorting on the DBMS)
        key = self._before_lookup(row, namemapping)
        if key is not None:
            return key

        self.targetconnection.execute(self.keyversionlookupsql, row,
                                      namemapping)

        versions = [kv for kv in self.targetconnection.fetchalltuples()]
        if not versions:
            # There is no existing version
            keyvalue = None
        else:
            # Look in all (key, version) pairs and find the key for the newest
            # version
            keyvalue, versionvalue = versions[-1]
            for kv in versions:
                if kv[1] > versionvalue:
                    keyvalue, versionvalue = kv

        self._after_lookup(row, namemapping, keyvalue)
        return keyvalue

    def scdensure(self, row, namemapping={}):
        """Lookup or insert a version of a slowly changing dimension member.

           .. Note:: Has side-effects on the given row.

           Arguments:
           - row: a dict containing the attributes for the member.
             key, versionatt, fromatt, and toatt are not required to be
             present but will be added (if defined).
           - namemapping: an optional namemapping (see module's documentation)
        """
        versionatt = (namemapping.get(self.versionatt) or self.versionatt)
        key = (namemapping.get(self.key) or self.key)
        if self.fromatt:  # this protects us against None in namemapping.
            fromatt = (namemapping.get(self.fromatt) or self.fromatt)
        else:
            fromatt = None
        if self.toatt:
            toatt = (namemapping.get(self.toatt) or self.toatt)
        else:
            toatt = None
        if self.srcdateatt:
            srcdateatt = (namemapping.get(self.srcdateatt) or self.srcdateatt)
        else:
            srcdateatt = None

        # Get the newest version and compare to that
        keyval = self.lookup(row, namemapping)

        if keyval is None:
            # It is a new member. We add the first version.
            row[versionatt] = 1
            if fromatt and fromatt not in row:
                if self.minfrom is not None:
                    # We need the following hack to distinguish between
                    # 'not set' and 'use the value None'...
                    if self.minfrom == (None,):
                        row[fromatt] = None
                    else:
                        row[fromatt] = self.minfrom
                else:
                    row[fromatt] = self.fromfinder(self.targetconnection,
                                                   row, namemapping)
            if toatt and toatt not in row:
                row[toatt] = self.maxto
            row[key] = self.insert(row, namemapping)
            return row[key]
        else:
            # There is an existing version. Check if the attributes are
            # identical
            type1updates = {}  # for type 1
            addnewversion = False  # for type 2
            other = self.getbykey(keyval)  # the full existing version
            for att in self.all:
                # Special (non-)handling of versioning and key attributes:
                if att in (self.key, self.versionatt):
                    # Don't compare these - we don't expect them to have
                    # meaningful values in row
                    continue
                # We may have to compare the "from dates"
                elif att == self.toatt:
                    if other[self.toatt] != self.maxto:
                        addnewversion = True

                elif att == self.fromatt:
                    if self.srcdateatt is None:  # We don't compare dates then
                        continue
                    else:
                        # We have to compare the dates in row[..] and other[..]
                        # We have to make sure that the dates are of comparable
                        # types.
                        rdt = self.srcdateparser(row[srcdateatt])
                        if rdt == other[self.fromatt]:
                            continue  # no change in the "from attribute"
                        elif isinstance(rdt, type(other[self.fromatt])):
                            # they are not equal but are of the same type, so
                            # we are dealing with a new date
                            addnewversion = True
                        else:
                            # They have different types (and are thus not
                            # equal). Try to convert to strings and see if they
                            # are equal.
                            modref = (self.targetconnection
                                      .getunderlyingmodule())
                            rowdate = modref.Date(rdt.year, rdt.month, rdt.day)
                            if str(rowdate).strip('\'"') != \
                                    str(other[self.fromatt]).strip('\'"'):
                                addnewversion = True
                # Handling of "normal" attributes:
                else:
                    mapped = (namemapping.get(att) or att)
                    if row[mapped] != other[att]:
                        if att in self.type1atts:
                            type1updates[att] = row[mapped]
                        else:
                            addnewversion = True
                if addnewversion and not self.type1atts:
                    # We don't have to look for possible type 1 updates
                    # and we already know that a type 2 update is needed.
                    break
                # else: continue

            if len(type1updates) > 0:
                # Some type 1 updates were found
                self.__performtype1updates(type1updates, other)

            if addnewversion:  # type 2
                # Make a new row version and insert it
                row.pop(key, None)
                row[versionatt] = other[self.versionatt] + 1
                if fromatt:
                    row[fromatt] = self.fromfinder(self.targetconnection,
                                                   row, namemapping)
                if toatt:
                    row[toatt] = self.maxto
                row[key] = self.insert(row, namemapping)
                # Update the todate attribute in the old row version in the DB.
                if toatt:
                    toattval = self.tofinder(self.targetconnection, row,
                                             namemapping)
                    self.targetconnection.execute(
                        self.updatetodatesql, {
                            self.key: keyval, self.toatt: toattval})
                # Only cache the newest version
                if self.__cachesize and keyval in self.rowcache:
                    del self.rowcache[keyval]
            else:
                # Update the row dict by giving version and dates and the key
                row[key] = keyval
                row[versionatt] = other[self.versionatt]
                if self.fromatt:
                    row[fromatt] = other[self.fromatt]
                if self.toatt:
                    row[toatt] = other[self.toatt]

            return row[key]

    def _before_lookup(self, row, namemapping):
        if self.__cachesize:
            namesinrow = [(namemapping.get(a) or a) for a in self.lookupatts]
            searchtuple = tuple([row[n] for n in namesinrow])
            return self.keycache.get(searchtuple, None)
        return None

    def _after_lookup(self, row, namemapping, resultkey):
        if self.__cachesize and resultkey is not None:
            namesinrow = [(namemapping.get(a) or a) for a in self.lookupatts]
            searchtuple = tuple([row[n] for n in namesinrow])
            self.keycache[searchtuple] = resultkey

    def _before_getbykey(self, keyvalue):
        if self.__cachesize:
            res = self.rowcache.get(keyvalue)
            if res is not None:
                return dict(zip(self.all, res))
        return None

    def _after_getbykey(self, keyvalue, resultrow):
        if self.__cachesize and resultrow[self.key] is not None:
            # if resultrow[self.key] is None, no result was found in the db
            self.rowcache[keyvalue] = tuple([resultrow[a] for a in self.all])

    def _before_update(self, row, namemapping):
        """ """
        # We have to remove old values from the caches.
        key = (namemapping.get(self.key) or self.key)
        for att in self.lookupatts:
            if (att in namemapping and namemapping[att] in row) or att in row:
                # A lookup attribute is about to be changed and we should make
                # sure that the cache does not map from the old value.  Here,
                # we can only see the new value, but we can get the old lookup
                # values by means of the key:
                oldrow = self.getbykey(row[key])
                searchtuple = tuple([oldrow[n] for n in self.lookupatts])
                if searchtuple in self.keycache:
                    del self.keycache[searchtuple]
                break

        if row[key] in self.rowcache:
            # The cached row is now incorrect. We must make sure it is
            # not in the cache.
            del self.rowcache[row[key]]

        return None

    def _after_insert(self, row, namemapping, newkeyvalue):
        """ """
        # After the insert, we can look it up. Pretend that we
        # did that. Then we get the new data cached.
        # NB: Here we assume that the DB doesn't change or add anything.
        # For example, a DEFAULT value in the DB or automatic type coercion can
        # break this assumption.
        # Note that we always cache inserted members (in CachedDimension
        # this is an option).
        if self.__cachesize:
            self._after_lookup(row, namemapping, newkeyvalue)
            tmp = pygrametl.project(self.all[1:], row, namemapping)
            tmp[self.key] = newkeyvalue
            self._after_getbykey(newkeyvalue, tmp)

    def __performtype1updates(self, updates, lookupvalues, namemapping={}):
        """ """
        # find the keys in the rows that should be updated
        self.targetconnection.execute(self.keylookupsql, lookupvalues,
                                      namemapping)
        updatekeys = [e[0] for e in self.targetconnection.fetchalltuples()]
        updatekeys.reverse()
        # Generate SQL for the update
        valparts = ", ".join(
            ["%s = %%(%s)s" % (self.quote(k), k) for k in updates])
        keyparts = ", ".join([str(k) for k in updatekeys])
        sql = "UPDATE %s SET %s WHERE %s IN (%s)" % \
              (self.name, valparts, self.quote(self.key), keyparts)
        self.targetconnection.execute(sql, updates)
        # Remove from our own cache
        for key in updatekeys:
            if key in self.rowcache:
                del self.rowcache[key]



import unittest
from simpleetl import FactTable, runETL, datatypes, CONFIG
import pg_creds
import psycopg2

SCHEMA = "testschema"
TABLE = "test_partitioning"
CONFIG.batch_size = 1
CONFIG.parallel_processes = 4
from simpleetl._functions._dbfunctions import table_exists, get_table_partitions

""" Test that ETL can be run, even though table should be partitioned but is not."""


def datagenerator(rangeadd=0):
    for n in range(100):
        n += rangeadd
        add_years, month = divmod(n, 12)
        month += 1
        year = 2010 + add_years
        yield {"lookupid": n, "datekey": "{y}{m}01".format(y=year, m=str(month).rjust(2, '0'))}


def partitionfunc_year(value):
    suffix = str(int(value / 10000))
    range_from = int("{y}0101".format(y=int(value / 10000)))
    range_to = int("{y}0101".format(y=int(value / 10000) + 1))

    return suffix, range_from, range_to


def etlrun(rangeadd, partition=False):
    f = FactTable(schema=SCHEMA, table=TABLE, lookupatts=["lookupid", "datekey"],
                  store_history=True)
    if partition:
        f.enable_partitioning("datekey", partitionfunc_year)
    f.handle_deleted_rows(method="wipe")
    f.add_column_mapping("lookupid", datatypes.bigint)

    f.add_column_mapping("datekey", datatypes.int_notnull)
    stats = runETL(f, datagenerator(rangeadd),
                   db_host=pg_creds.PG_HOST, db_port=pg_creds.PG_PORT,
                   db_name=pg_creds.PG_DBNAME, db_user=pg_creds.PG_USER,
                   db_pass=pg_creds.PG_PASS)
    return stats


def cleanup():
    dsn = ("dbname={dbname} user={username} password={password} host={host} port={port}"
           .format(dbname=pg_creds.PG_DBNAME, username=pg_creds.PG_USER,
                   password=pg_creds.PG_PASS, host=pg_creds.PG_HOST, port=pg_creds.PG_PORT))
    dbcon = psycopg2.connect(dsn)
    with dbcon:
        with dbcon.cursor() as cursor:
            cursor.execute("DROP SCHEMA IF EXISTS " + SCHEMA + " CASCADE")
    dbcon.close()
    print("Cleaned up")


def get_partitions():
    dsn = ("dbname={dbname} user={username} password={password} host={host} port={port}"
           .format(dbname=pg_creds.PG_DBNAME, username=pg_creds.PG_USER,
                   password=pg_creds.PG_PASS, host=pg_creds.PG_HOST, port=pg_creds.PG_PORT))
    dbcon = psycopg2.connect(dsn)
    partitions = get_table_partitions(dbcon, SCHEMA, TABLE)
    dbcon.close()
    return partitions


class TestHistoricLoad(unittest.TestCase):

    def test_partitioning(self):
        etlrun(0, False)
        # test = get_partitions()
        # self.assertListEqual(test, [TABLE + "_" + str(y) for y in range(2010, 2019)])

        print(" NEXT RUN ")

        etlrun(0, True)
        # test = get_partitions()
        # self.assertListEqual(test, [TABLE + "_" + str(y) for y in range(2014, 2023)])

    def setUp(self):
        cleanup()

    def tearDown(self):
        cleanup()


if __name__ == '__main__':
    unittest.main()



import logging
import psycopg2


## Logging handler for PostgreSQL
#
#
class pgLogHandler(logging.Handler):
    __schema = "simpleetl"
    __table = "log"
    initial_sql = """CREATE SCHEMA IF NOT EXISTS {schema};
                    CREATE TABLE IF NOT EXISTS {schema}.{table}(
                    logid bigserial,
                        logtime timestamp with time zone DEFAULT NOW(),
                        batchkey int,
                        created text,
                        name text,
                        loglevel int,
                        loglevelname text,
                        message text,
                        args text,
                        module text,
                        funcname text,
                        lineno int,
                        exception text,
                        process int,
                        thread text,
                        threadname text
                   )""".format(schema=__schema, table=__table)

    insertion_sql = """INSERT INTO {schema}.{table}(
                            batchkey,
                            created,
                            name,
                            loglevel,
                            loglevelname,
                            message,
                            module,
                            funcname,
                            lineno,
                            exception,
                            process,
                            thread,
                            threadname) VALUES (
                            %(batchkey)s,
                            %(created)s,
                            %(name)s,
                            %(levelno)s,
                            %(levelname)s,
                            %(msg)s,
                            %(module)s,
                            %(funcName)s,
                            %(lineno)s,
                            %(exc_text)s,
                            %(process)s,
                            %(thread)s,
                            %(threadName)s
                    )""".format(schema=__schema, table=__table)

    def connect(self, db_host, db_name, db_user, db_pass, db_port):

        try:

            self.__connect = psycopg2.connect(
                database=db_name,
                host=db_host,
                user=db_user,
                password=db_pass,
                port=db_port)
            self.__connect.autocommit = True

            return True
        except:
            return False

    def __init__(self, lbatchkey, db_host, db_name, db_user, db_pass, db_port):

        self.__batchkey = -1

        self.__connect = None

        if not self.connect(db_host, db_name, db_user, db_pass, db_port):
            print("Database connection error, no logging")
            return

        logging.Handler.__init__(self)

        self.__connect.cursor().execute(pgLogHandler.initial_sql)
        self.__connect.commit()
        self.__connect.cursor().close()

    def connect_batch(self, batchkey):
        self.__batchkey = batchkey

    def emit(self, record):

        # Use default formatting:
        self.format(record)

        if record.exc_info:
            record.exc_text = logging._defaultFormatter.formatException(record.exc_info)
        else:
            record.exc_text = ""
        recdict = record.__dict__
        recdict["batchkey"] = self.__batchkey

        try:
            cur = self.__connect.cursor()
        except:
            self.connect()
            cur = self.__connect.cursor()

        cur.execute(pgLogHandler.insertion_sql, recdict)
        self.__connect.commit()
        self.__connect.cursor().close()



import datetime
from math import ceil

from pygrametl import getvalue

from simpleetl import Dimension
from simpleetl import datatypes

# Define a "default" date, when no date is found.
nulldate = {}
nulldt = -1
nulldate["date"] = None
nulldate["holiday"] = False
for r in ["datekey", "year", "month", "day", "quarter", "weekday", "iso_year", "iso_weeknumber",
          "iso_weekday", "month_str", "day_us_str", "season", "season_us_str"]:
    nulldate[r] = -1


def _datehandling(row, namemapping):
    """ A method for handling and converting date formats from a string and
        exploding and adding it to a dictionary.

    :param row: The data dictionary, which contains the raw date, but which
                the output is going to be applied to too.
    :param namemapping: A dictionary describing key mappings of the input
                        dictionary.
    :return: A dictionary, containing values for the dimdate dimension.
    """
    datekey = getvalue(row, 'datekey', namemapping)

    if datekey is None or datekey == nulldt:
        return nulldate
    dt = datetime.datetime.strptime(str(datekey), "%Y%m%d", )
    outrow = {}

    outrow['day'] = dt.day
    outrow['month'] = dt.month
    outrow['year'] = dt.year
    outrow['quarter'] = ceil(dt.month / 3)
    outrow['weekday'] = dt.weekday()
    (outrow["iso_year"],
     outrow['iso_weeknumber'],
     outrow['iso_weekday']) = dt.isocalendar()
    outrow['datekey'] = dt.year * 10000 + dt.month * 100 + dt.day
    outrow['date'] = dt.strftime("%Y-%m-%d")
    wds = ["Mandag", "Tirsdag", "Onsdag",
           "Torsdag", "Fredag", "Lørdag", "Søndag"]

    mrds = ["Januar", "Februar", "Marts", "April", "Maj", "Juni", "Juli", "August", "September", "Oktober", "November",
            "December"]
    outrow["month_str"] = mrds[dt.month - 1]
    outrow["day_us_str"] = wds[dt.weekday()]
    outrow["holiday"] = False
    # import holidays
    # dk_holidays = holidays.Denmark()
    # dt in dk_holidays or (dt.month == 12 and dt.day in (24, 31))
    season = 0
    if 3 <= dt.month <= 5:
        season = 1
    elif 6 <= dt.month <= 8:
        season = 2
    elif 9 <= dt.month <= 11:
        season = 3
    seasons = ["Vinter", "Forår", "Sommer", "Efterår"]
    outrow["season"] = season
    outrow["season_us_str"] = seasons[season]

    return outrow


dimdate = Dimension(schema="testschema", table="dimdate", key="datekey",
                    rowexpander=_datehandling, integerkey=True)

dimdate.add_lookupatt("datekey", datatypes.datekey, nulldt)

dimdate.add_att('year', datatypes.smallint)
dimdate.add_att('month', datatypes.smallint)
dimdate.add_att('day', datatypes.smallint)
dimdate.add_att('weekday', datatypes.smallint)
dimdate.add_att('iso_weeknumber', datatypes.smallint)
dimdate.add_att('iso_weekday', datatypes.smallint)
dimdate.add_att('iso_year', datatypes.smallint)
dimdate.add_att('quarter', datatypes.smallint)

dimdate.add_att('season_us_str', datatypes.text)
dimdate.add_att('day_us_str', datatypes.text)
dimdate.add_att('month_str', datatypes.text)

dimdate.add_att("date", datatypes.date)
dimdate.add_att("holiday", datatypes.boolean)



import unittest
import pg_creds
import psycopg2
from examples.load_user_example import load_user_etl

SCHEMA = "testschema"


def cleanup():
    dsn = ("dbname={dbname} user={username} password={password} host={host} port={port}"
           .format(dbname=pg_creds.PG_DBNAME, username=pg_creds.PG_USER,
                   password=pg_creds.PG_PASS, host=pg_creds.PG_HOST, port=pg_creds.PG_PORT))
    dbcon = psycopg2.connect(dsn)
    with dbcon:
        with dbcon.cursor() as cursor:
            cursor.execute("DROP SCHEMA IF EXISTS " + SCHEMA + " CASCADE")
    dbcon.close()
    print("Cleaned up")


class TestExamples(unittest.TestCase):

    def test_load_user(self):
        stats_expected = {'facts_inserted': 4, 'facts_updated': 0, 'facts_deleted': 0, 'historic_inserted': 0,
                        'historic_updated': 0, 'historic_deleted': 0, 'rows_read': 4, 'rows_parsed': 4}
        stats = load_user_etl()

        stats_parsed = {key: stats[key] for key in stats_expected.keys()}
        self.assertDictEqual(stats_expected, stats_parsed)

    def setUp(self):
        cleanup()

    def tearDown(self):
        cleanup()


if __name__ == '__main__':
    unittest.main()



from simpleetl import LOG

from simpleetl._functions import _datatypes
from simpleetl._functions._db_quoting import quotelist, quote
from simpleetl._functions._dbfunctions import has_index, table_exists, column_type, has_fkey, \
    create_fkey, has_unique, \
    create_unique


def _ensure_lookupindex(pgcon, schema, table, cols, historic=False, unique=True):
    """

    :param dbcon:
    :param schema:
    :param table:
    :param hist:
    :return:
    """

    uniquesql = ""
    if unique:
        uniquesql = "UNIQUE"
    org_cols = None
    if historic:
        org_cols = cols.copy()
        cols.append("_validfrom")
        cols.append("_validto")
        cols.append("_version")

    if not has_index(pgcon, schema, table, cols):
        with pgcon.cursor() as cursor:
            cursor.execute("""CREATE {unique} INDEX ON {schema}.{table}
                        ({keycols});
                        """.format(schema=quote(schema, pgcon),
                                   table=quote(table, pgcon),
                                   unique=uniquesql,
                                   keycols=", ".join(quotelist(cols, pgcon))))
    # if historic:
    #    dbcon.dbexecute("""ALTER TABLE {schema}.{table} ADD UNIQUE ({keycols});"""
    #                    .format(schema=quote(schema, dbcon),
    #                            table=quote(table, dbcon),
    #                               keycols=", ".join(quotelist(org_cols, dbcon)))) 


def _ensure_structure(pgconn, factobj):
    factcols = factobj._get_cols_types(include_changeatts=True, include_ignorecols=True)
    partition_att = None
    if factobj.partitioning_enabled:
        partition_att = factobj.partition_attribute
    create_tab = _ensure_facttable_schema(pgconn, factobj.schema, factobj.table, factobj.key, factobj.keytype,
                                          factcols, partition_att=partition_att)
    lookupcols = factobj._get_cols(only_lookupatts=True)
    keycols = lookupcols.copy()
    if keycols is not None and keycols != [factobj.key]:
        # if keycols == [factobj.key] the primary key is the lookup attribute
        if not factobj.key_in_atts:
            keycols.append(factobj.key)

        _ensure_lookupindex(pgconn, factobj.schema, factobj.table, keycols)

        if lookupcols:
            if not has_unique(pgconn, factobj.schema, factobj.table, lookupcols):
                LOG.debug("Creating unique constraint on fact table")
                create_unique(pgconn, factobj.schema, factobj.table, lookupcols)

    if factobj.store_history:
        # factid = "_" + factobj.table + "_" + factobj.key
        partition_att_hist = None
        if factobj.partitioning_enabled_hist:
            partition_att_hist = factobj.partition_attribute_hist
        histkeycols = []
        if keycols:
            histkeycols += keycols.copy()
        # histkeycols.append(factid)
        addcols = [("_validfrom", _datatypes.timestamp_with_timezone),
                   ("_validto", _datatypes.timestamp_with_timezone),
                   ("_version", _datatypes.int)]

        histcols = factobj._get_cols_types(include_changeatts=False, include_ignorecols=True)
        create_hist = _ensure_facttable_schema(pgconn, factobj.schema, factobj.table_historic,
                                               factobj.key, factobj.keytype, histcols
                                               , addcols, partition_att=partition_att_hist)

        if keycols:
            _ensure_lookupindex(pgconn, factobj.schema, factobj.table_historic, keycols, True)
            lookupcols_validfrom = lookupcols.copy()
            lookupcols_validfrom.append("_validfrom")
            lookupcols_version = lookupcols.copy()
            lookupcols_version.append("_version")
            if not has_unique(pgconn, factobj.schema, factobj.table_historic, lookupcols_validfrom):
                LOG.debug("Creating unique constraint on historic table including _validfrom")
                create_unique(pgconn, factobj.schema, factobj.table_historic, lookupcols_validfrom)
            if not has_unique(pgconn, factobj.schema, factobj.table_historic, lookupcols_version):
                LOG.debug("Creating unique constraint on historic table including _version")
                create_unique(pgconn, factobj.schema, factobj.table_historic, lookupcols_version)

    _ensure_dimensions(pgconn, factobj.schema, factobj.table, factobj.dim_mappings)
    if factobj.store_history:
        _ensure_dimensions(pgconn, factobj.schema, factobj.table_historic, factobj.dim_mappings)
    pgconn.commit()


def _ensure_facttable_schema(pgconn, schema, table, key, keytype, columns, histcols=None,
                             partition_att=None):
    create_table = False
    partatt = ""
    partrange = ""
    prikey = quote(key, pgconn)
    if not table_exists(pgconn, schema, table):
        if partition_att is not None:
            partrange = "PARTITION BY RANGE ({c})".format(c=partition_att)
            if key != partition_att:
                # Postgresql requires partition key to be part of primart key
                prikey += ", " + quote(partition_att, pgconn)
            for col, datatype in columns:
                if col == partition_att:
                    partatt = "{n} {t},".format(n=quote(col, pgconn), t=datatype.sqltype())
                    break

        with pgconn.cursor() as cursor:
            cursor.execute("CREATE SCHEMA if not exists {sch};"
                           .format(sch=quote(schema, pgconn)))
            create_table = True
            tabcreate = """CREATE TABLE if not exists {sch}.{tab}
                ({k} {keytype} NOT NULL,
                {partatt}
            PRIMARY KEY ({prikey})
           ) {partrange}""".format(sch=quote(schema, pgconn),
                                   tab=quote(table, pgconn),
                                   k=quote(key, pgconn),
                                   prikey=prikey,
                                   keytype=keytype.sqltype(),
                                   partatt=partatt,
                                   partrange=partrange)
            cursor.execute(tabcreate)

    if histcols is not None:
        columns += histcols

    for col, datatype in columns:
        addcol(schema, table, pgconn, col, datatype)
    return create_table


def _ensure_dimensions(pgconn, schema, table, dim_mappings):
    for dim in dim_mappings:
        if not dim.fkey or not dim.insert:
            continue
        dstschema = dim.dimension._get_schema()
        dsttbl = dim.dimension._get_table()
        dstcol = dim.dimension._get_key()
        tablecol = dim.dstcol
        exists = has_fkey(pgconn, schema, table, tablecol, dstschema, dsttbl, dstcol)
        if not exists:
            try:
                create_fkey(pgconn, schema, table, tablecol, dstschema, dsttbl,
                            dstcol, True)
            except Exception as err:
                LOG.error("Problem creating foreign key {s1}.{t1}.{c1} -> {s2}.{t2}.{c2}"
                          .format(s1=schema, t1=table, c1=tablecol,
                                  s2=dstschema, t2=dsttbl, c2=dstcol))


def _create_temp_table(pgconn, factobj):
    # self.__lookupatts
    """Create temporary data storage, for loading data into."""

    with pgconn.cursor() as cursor:
        cursor.execute("CREATE SCHEMA IF NOT EXISTS tmp_data")

        cursor.execute(
            "DROP TABLE IF EXISTS tmp_data.{tmptable}".format(
                tmptable=quote(factobj.tmptable, pgconn)))
        factcols = factobj._get_cols_types(include_changeatts=True, include_ignorecols=True)
        create_tmp_tab = _ensure_facttable_schema(pgconn, "tmp_data", factobj.tmptable, factobj.key,
                                                  factobj.keytype, factcols)
        # When creating temporary table, we do not use a primary key, unless it is part of lookup attributes
        if not factobj.lookupatts or factobj.key not in factobj.lookupatts:
            cursor.execute(
                "ALTER TABLE tmp_data.{tmptable} DROP COLUMN {key};".format(key=quote(factobj.key, pgconn),
                                                                            tmptable=quote(factobj.tmptable,
                                                                                           pgconn)))


def _drop_temp_table(pgconn, factobj):
    """Create temporary data storage, for loading data into."""
    q = """DROP TABLE IF EXISTS tmp_data.{tmptable};
           """.format(schema=quote(factobj.schema, pgconn), tmptable=quote(factobj.tmptable, pgconn))
    with pgconn.cursor() as cursor:
        cursor.execute(q)


def _ensure_indexes(pgconn, factobj):
    idx = """CREATE INDEX ON {schema}.{table}
                            ({keycols});"""

    for idxcols in factobj.indexes:
        if not has_index(pgconn, factobj.schema, factobj.table, idxcols):
            LOG.debug("Creating index on columns: " + ",".join(idxcols))
            with pgconn.cursor() as cursor:
                cursor.execute(idx.format(schema=quote(factobj.schema, pgconn),
                                          table=quote(factobj.table, pgconn),
                                          keycols=", ".join(quotelist(idxcols, pgconn))))
        # if (factobj.store_history and
        #        not dbcon.has_index(factobj.schema, factobj.table_historic, idxcols)):
        #    dbcon.dbexecute(idx.format(schema=quote(factobj.schema, dbcon),
        #                               table=quote(factobj.table_historic, dbcon),
        #                               keycols=", ".join(quotelist(idxcols, dbcon))))


def addcol(schema, table, pgconn, col, datatype):
    coltype = column_type(pgconn, schema, table, col)
    if coltype is not None:
        if coltype.replace(" NOT NULL", "").upper() != datatype.sqltype().replace(" NOT NULL", "").upper():
            LOG.warning("Wrong datatype ({tnow}). Consider running ALTER TABLE {s}.{t} ALTER COLUMN {c} TYPE {ty};"
                        .format(s=schema, t=table, c=col,
                                ty=datatype.sqltype().replace(" NOT NULL", ""),
                                tnow=coltype.replace(" NOT NULL", "")))
        # Test if current datatype allows NULL and whether NULLs should be allowed
        if ("NOT NULL" not in coltype.upper()) != datatype.allow_null:
            setval = "SET NOT NULL"
            if datatype.allow_null:
                setval = "DROP NOT NULL"
            q = "ALTER TABLE {s}.{t} ALTER COLUMN {c} {set}".format(s=schema, t=table, c=col, set=setval)
            LOG.warning("Wrong NULL settings. Consider running " + q)
    else:
        null = ""

        if not datatype.allow_null:
            null = " NOT NULL"
        with pgconn.cursor() as cursor:
            cursor.execute('ALTER TABLE {sc}.{tab} ADD COLUMN {n} {t} {null}'
                           .format(sc=quote(schema, pgconn),
                                   tab=quote(table, pgconn),
                                   n=quote(col, pgconn),
                                   t=datatype.sqltype(),
                                   null=null))



import sys
from simpleetl import Dimension
from simpleetl import LOG
from simpleetl._functions import _datatypes as dt
from simpleetl._modules.Dimension import Column, DimColumn
from simpleetl import datatypes
from functools import lru_cache


class FactTable:
    def __init__(self, schema, table,
                 lookupatts=None,
                 store_history=False,
                 migrate_updates=True,
                 key=None,
                 ignore_atts=None,
                 track_last_updated=False,
                 track_created=False,
                 tmptable=None):
        """
        :param schema: Database schema for fact table
        :param table: Database table for fact table
        :param lookupatts: A list of attributes which uniquely defines a row. Will enable handling of changes to facts.
        :param store_history: Will store a complete track of changes to facts in a seperate table.
        :param migrate_updates: Update facts when changes are deteted in source data.
        :param key: Primary key name.
        :param ignore_atts: List of attributes, where changes will be ignore.
        :param track_last_updated: Add an attribute for tracking last changed timestamp of a fact.
        :param track_created: Add an attribute for tracking when a fact is created.
        :param tmptable: Temporary table, stored in the tmp_data schema. Lock on this table will ensure only single ETL
        on this table at a time. If none, generated from schema and table name.
        """
        self.__schema = schema
        self.__table = table
        self.__store_history = store_history
        self.__lookupatts = None
        self.__ignore_atts = []
        self.__track_last_updated = track_last_updated
        self.__track_created = track_created
        self.__migrate_updates = migrate_updates
        if ignore_atts:
            self.__ignore_atts = ignore_atts.copy()
        if lookupatts:
            assert isinstance(lookupatts, list), "Lookup attributs must be a list"
            self.__lookupatts = lookupatts.copy()  # Make a copy of the list using .copy(), as the list is parsed by
        elif store_history:
            LOG.error("History cannot be preserved when no lookupatts have been defined")
            sys.exit(1)

        # reference. If we didn't do this, the caller could later change the lookupatts.

        self.__table_historic = None
        self.__deleted_rows_method = False
        self.__handle_deleted_rows_func = None
        self.__deleted_mark_keep_days = None
        self.__deleted_rows_limit = None

        self.__key = key == None and "id" or key  # set key to 'id' if key is None
        self.__keytype = datatypes.bigint_notnull
        self.__key_in_atts = False
        self.__ignore_pk_in_atts_warning = False

        if self.__store_history:
            self.__table_historic = self.__table + "_historic"

        self.__simple_mappings = []
        self.__dim_mappings = []

        self.__indexes = []

        self.__prepared = False

        self.__col_order = []
        if tmptable:
            self.__tmptable = "_".join([self.__schema, tmptable, "tmp"])
        else:
            self.__tmptable = "_".join([self.__schema, self.__table, "tmp"])

        self.__partition_attribute = None
        self.__partitionfunc = None
        self.__partitions_used = set()

        self.__partition_attribute_hist = None
        self.__partitionfunc_hist = None
        self.__partitions_used_hist = set()

    @property
    def schema(self):
        return self.__schema

    @property
    def table(self):
        return self.__table

    @property
    def tmptable(self):
        return self.__tmptable

    @property
    def table_historic(self):
        return self.__table_historic

    @property
    def store_history(self):
        return self.__store_history

    @property
    def track_last_updated(self):
        return self.__track_last_updated

    @property
    def track_created(self):
        return self.__track_created

    @property
    def migrate_updates(self):
        return self.__migrate_updates

    @property
    def lookupatts(self):
        return self.__lookupatts and self.__lookupatts.copy() or None

    # def add_lookupatt(self, att):
    #    self.__lookupatts.append(att)

    @property
    def deleted_rows_method(self):
        return self.__deleted_rows_method

    @property
    def handle_deleted_rows_func(self):
        return self.__handle_deleted_rows_func

    @property
    def deleted_mark_keep_days(self):
        return self.__deleted_mark_keep_days

    @property
    def deleted_rows_limit(self):
        return self.__deleted_rows_limit

    @property
    def key(self):
        return self.__key

    @property
    def keytype(self):
        return self.__keytype

    @property
    def simple_mappings(self):
        return self.__simple_mappings

    @property
    def dim_mappings(self):
        return self.__dim_mappings

    @property
    def indexes(self):
        return self.__indexes

    @property
    def key_in_atts(self):
        return self.__key_in_atts

    @property
    def ignore_pk_in_atts_warning(self):
        return self.__ignore_pk_in_atts_warning

    """ Properties and methods related to table partitioning """

    @property
    def partitioning_enabled(self):
        return self.__partition_attribute is not None

    @property
    def partitioning_enabled_hist(self):
        return self.__partition_attribute_hist is not None

    @property
    def partition_attribute(self):
        return self.__partition_attribute

    @property
    def partition_attribute_hist(self):
        return self.__partition_attribute_hist

    def get_partition(self, row):
        return self.__get_partition_att(row[self.__partition_attribute])

    def get_partition_hist(self, row):
        return self.__get_partition_att_hist(row[self.__partition_attribute_hist])

    @lru_cache(1000)
    def __get_partition_att(self, value):
        return self.__partitionfunc(value)

    @lru_cache(1000)
    def __get_partition_att_hist(self, value):
        return self.__partitionfunc_hist(value)

    def add_used_partition(self, partition):
        self.__partitions_used.add(partition)

    def add_used_partition_hist(self, partition):
        self.__partitions_used_hist.add(partition)

    @property
    def used_partitions(self):
        plist = list(self.__partitions_used)
        plist.sort(key=lambda x: x[0])
        return plist

    @property
    def used_partitions_hist(self):
        plist = list(self.__partitions_used_hist)
        plist.sort(key=lambda x: x[0])
        return plist

    def enable_partitioning(self, attribute, partitionfunc):
        """Enables range parition on fact table.
        :param attribute: Attribute, which must be part of lookup attributes, to partition by.
        :param partitionfunc: Function, which takes an input value and returns a tuple of 3 values:
            (table_suffix, range_start, range_end). Example: A table partitioned on a date column with partitions per year
            could return on input 2020-03-10: ('2020', '2020-01-01', '2021-01-01').
            * PLEASE NOTE, the range_end is exclusive.
            * The function will be executed wrapped in a cache, to prevent expensive calculations to be repeated.
        """

        assert attribute in self.__lookupatts, (
                "Partition attribute " + str(attribute) + " must be part of lookup attributes on table, as PostgreSQL"
                                                          " requires partition key to be part of unique indexes.")
        self.__partition_attribute = attribute
        self.__partitionfunc = partitionfunc

    def enable_partitioning_historic(self, attribute, partitionfunc):
        """Enables range parition on historic fact table.
        :param attribute: Attribute, which must be part of lookup attributes, to partition by.
        :param partitionfunc: Function, which takes an input value and returns a tuple of 3 values:
            (table_suffix, range_start, range_end). Example: A table partitioned on a date column with partitions per year
            could return on input 2020-03-10: ('2020', '2020-01-01', '2021-01-01').
            * PLEASE NOTE, the range_end is exclusive.
            * The function will be executed wrapped in a cache, to prevent expensive calculations to be repeated.
        """

        assert attribute in self.__lookupatts, (
                "Partition attribute " + str(
            attribute) + " on historic table must be part of lookup attributes on table, as PostgreSQL"
                         " requires partition key to be part of unique indexes.")
        self.__partition_attribute_hist = attribute
        self.__partitionfunc_hist = partitionfunc

    def set_ignore_pk_in_atts_warning(self):
        """Silense warnings when adding attribute matching primary key"""
        self.__ignore_pk_in_atts_warning = True

    def add_dim_mapping(self, dimension, dstcol, namemapping={}, inject=False, insert_to_facttable=True,
                        fkey=True, index=True, allow_null=False):
        """Adds a dimension to the fact table.

        :param dimension: Dimension object
        :param dstcol: Destination column of fact table
        :param namemapping: Optional dictinary mapping for dimension. E.g., if dimension expects "datekey" but we have
        "invoicedate" from source, namemapping would be {"datekey":"invoicedate"}
        :param inject: Inject found key/val into src data, e.g., for use with other dimension lookups.
        :param insert_to_facttable: Insert dimension key into fact table if True.
        :param fkey: Create foreing key to dimension.
        :param index: Create index on column.
        :param allow_null: Allow null values to column
        """
        if not isinstance(dimension, Dimension):
            LOG.error(
                "add_dim_mapping " + dstcol + " dimension must be a Dimension() object. Got " + str(type(dimension)))
            sys.exit(1)

        if not isinstance(dstcol, str) or not dstcol:
            LOG.error("add_column_mapping dstcol must be a string. Input value: " + str(dstcol))
            sys.exit(1)

        if not isinstance(namemapping, dict):
            LOG.error("add_column_mapping " + dstcol + " namemapping must be a dictionary. Input type: " + str(
                type(namemapping)))
            sys.exit(1)

        for tp, k in self.__col_order:
            if k.dstcol == dstcol:
                LOG.error("add_column_mapping column allready added before. Column: " + str(dstcol))
                sys.exit(1)

        if not inject and not insert_to_facttable:
            LOG.error("Neither injecting dimension" + dimension._get_full_table() + "/" + dstcol +
                      " into source data or insertering into table, does not make use of dimension at all.")
            sys.exit(1)

        if not isinstance(fkey, bool):
            LOG.error("fkey must be boolean. Found " + str(fkey) + " of type " + str(type=fkey))
            sys.exit(1)
        if not isinstance(fkey, bool):
            LOG.error("index must be boolean. Found " + str(index) + " of type " + str(type=index))
            sys.exit(1)
        if dstcol == self.__key and not self.__ignore_pk_in_atts_warning:
            LOG.warning(dstcol + " is the same as primary key")
            self.__key_in_atts = True
            self.__keytype = dimension._get_keytype()
        if dstcol == self.__partition_attribute and allow_null:
            LOG.error("column " + dstcol + " is partition key and therefore must not be null.")
            sys.exit(1)
        if dstcol == self.__partition_attribute_hist and allow_null:
            LOG.error("column " + dstcol + " is partition key on historic table and therefore must not be null.")
            sys.exit(1)

        d = DimColumn(dimension, dstcol, namemapping, inject, insert_to_facttable, fkey, allow_null)

        self.__dim_mappings.append(d)
        self.__col_order.append(("d", d))
        if insert_to_facttable and index:
            self.add_index([dstcol])
        return

    def add_column_mapping(self, srccol, datatype, dstcol=None):
        """ Maps a value from source data to destination table.
        :param srccol: Source data column to use.
        :param datatype: Datatype of source data.
        :param dstcol: Name of destination column.
        """
        if not isinstance(srccol, str) or not srccol:
            LOG.error("add_column_mapping srccol must be a string. Input value: " + str(srccol))
            sys.exit(1)
        if dstcol is not None and not isinstance(dstcol, str):
            LOG.error("add_column_mapping dstcol must be a string. Input value: " + str(dstcol))
            sys.exit(1)

        if dstcol is None:
            dstcol = srccol

        for tp, k in self.__col_order:
            if k.dstcol == dstcol:
                LOG.error("add_column_mapping column allready added before. Column: " + str(dstcol))
                sys.exit(1)

        if dstcol == self.__partition_attribute and datatype.allow_null:
            LOG.error("column " + dstcol + " is partition key and therefore must not be null.")
            sys.exit(1)

        if dstcol == self.__partition_attribute_hist and datatype.allow_null:
            LOG.error("column " + dstcol + " is partition key on historic table and therefore must not be null.")
            sys.exit(1)

        if dstcol == self.__key and not self.__ignore_pk_in_atts_warning:
            LOG.warning(dstcol + " is the same as primary key")
            self.__key_in_atts = True
            datatype = datatype.copy(False)  # Ensure we disallow NULL
            self.__keytype = datatype

        if self.__lookupatts and dstcol in self.__lookupatts:
            datatype.copy(False)  # Ensure we disallow NULL
        c = Column(srccol, dstcol, datatype)
        self.__simple_mappings.append(c)
        self.__col_order.append(("c", c))

    def handle_deleted_rows(self, method="date", limitfunc=None, keep_days=None,
                            limit_deleted_rows=None):
        """Propagates deleted rows from source file to database.
        :param method: Determine how to handle deletes:
            "date": Mark deleted row in versioned table with an time stamp to _validto column. Deletes from T1 fact table.
            "mark": Add additional _deleted measure to fact table and versioned table. Will also mark _validto with end of life time stamp.
            "wipe": Completely wipes removed data from both T1 fact table and versioned fact table.
        :param limitfunc: a function returning a a SQL WHERE limit clause to constrain which rows will be deleted.
        :param keep_days: If set to an integer, delete marked data after keep_days days + 6 hours.
        :param limit_deleted_rows: Raise an error if more than n rows are deleted in a batch. Applies for non-historic table.
        """
        if self.lookupatts is None:
            LOG.error("Deleted rows cannot be handled when no lookupatts are defined.")
            sys.exit(1)
        if method not in ("date", "mark", "wipe"):
            LOG.error("Invalid value for handle_deleted_rows method: " + str(method))
            sys.exit(1)
        self.__deleted_rows_method = method
        self.__handle_deleted_rows_func = limitfunc
        if isinstance(keep_days, int):

            assert method == "mark", (
                    "When keep_days is set, method must be mark. Defined method is " + str(method))
            self.__deleted_mark_keep_days = int(keep_days)

        if limit_deleted_rows is not None:
            assert isinstance(limit_deleted_rows, int), (
                    "limit_deleted_rows must be an integer. Found " + str(type(limit_deleted_rows)))
            assert limit_deleted_rows > 0, (
                    "limit_deleted_rows must be greater than 0. Got " + str(limit_deleted_rows))
            self.__deleted_rows_limit = limit_deleted_rows

    def add_index(self, columns):
        """ Adds an index over a list of columns. This index is will be checked for existance at every ETL run.

        :param columns: List of columns to index in one index.
        """
        if not isinstance(columns, list):
            LOG.error("Index list must be list!")
            sys.exit(1)
        self.__indexes.append(columns)

    def _get_cols_types(self, dims=True, atts=True, include_lookupatts=True, only_lookupatts=False,
                        use_srccol=False, include_ignorecols=False, include_changeatts=False):
        """

        :param dims:
        :param atts:
        :param include_lookupatts:
        :param only_lookupatts:
        :param use_srccol:
        :param include_ignorecols:
        :param include_changeatts:
        :return:
        """
        cols = []
        for c_type, c_elem in self.__col_order:
            if dims and c_type == "d":  # Handle dimension mapping

                if not c_elem.insert:
                    continue  # Do not insert into table, dimension value only for looking up data
                if not include_ignorecols and c_elem.dstcol in self.__ignore_atts:
                    continue  # Ignore changes to this dimension attribute
                if use_srccol and len(c_elem.namemapping) > 0:

                    for nkey, n in c_elem.namemapping.items():
                        # nsrc = c.namemapping.get(nkey, n)

                        cdatatype = c_elem.dimension.get_lookuptype(nkey).copy(allow_null=c_elem.allow_null)

                        cols.append((n, cdatatype))
                elif use_srccol:
                    for srccol, srctype, defval in c_elem.dimension.get_lookupatts():
                        cols.append((srccol, srctype))
                else:

                    cdatatype = c_elem.dimension._get_keytype().copy(allow_null=c_elem.allow_null)
                    cols.append((c_elem.dstcol, cdatatype))
            elif atts and c_type == "c":  # Handle attribute mapping
                if not include_ignorecols and c_elem.dstcol in self.__ignore_atts:
                    continue  # Ignore changes to this attribute
                if include_ignorecols and not include_changeatts and c_elem.dstcol in ("_updated", "_deleted"):
                    continue  # Ignore last updated attribute
                if use_srccol:
                    cols.append((c_elem.srccol, c_elem.datatype))
                else:
                    cols.append((c_elem.dstcol, c_elem.datatype))

        # Remove lookup attributes, if we do not want these.
        if not include_lookupatts and self.lookupatts:
            cols2 = []
            for c in cols:
                if c[0] not in self.lookupatts:
                    cols2.append(c)
            cols = cols2

        if only_lookupatts:
            if self.lookupatts:
                c2 = []
                for c in cols:
                    if c[0] in self.lookupatts:
                        c2.append(c[0])
                if len(c2) != len(self.lookupatts):
                    errstr = "Lookup attribute missing: Defined colums: (" + str(c2) + "). Expected columns: " + str(
                        self.lookupatts)
                    LOG.error(errstr)
                    raise AttributeError(errstr)
                cols = [(c,) for c in self.lookupatts]
            else:
                cols = []

        return cols

    def _get_cols(self, dims=True, atts=True, include_lookupatts=True, only_lookupatts=False, use_srccol=False,
                  include_ignorecols=False, include_changeatts=False):
        """

        :param dims:
        :param atts:
        :param include_lookupatts:
        :param only_lookupatts:
        :param use_srccol:
        :param include_ignorecols:
        :param include_changeatts:
        :return:
        """
        cols = self._get_cols_types(dims, atts, include_lookupatts, only_lookupatts, use_srccol, include_ignorecols,
                                    include_changeatts)
        cols = [c[0] for c in cols]
        return cols

    def _get_srccols(self):
        cols = []
        injectcols = []
        for c in self.__dim_mappings:
            if not c.insert:
                continue
            for srccol, srctype, defval in c.dimension.get_lookupatts():

                cols.append(c.namemapping.get(srccol, srccol))
                if c.inject:
                    injectcols.append(c.dstcol)
        cols += [c.srccol for c in self.__simple_mappings]
        cols = set(cols)
        for c in injectcols:
            cols.remove(c)

        return list(cols)

    def _validate_structure(self):
        latts = self._get_cols(only_lookupatts=True)
        if latts and self.lookupatts:
            for latt in self.lookupatts:
                if latt not in latts:
                    LOG.error('Lookup attribute "' + latt + '" is not defined as an attribute on the fact table')
                    sys.exit(1)

    def _prepare_run(self):
        if self.__prepared:
            LOG.error("Fact table prepare_run has allready been executed once!")
            sys.exit(1)
        self.__prepared = True
        if self.__track_last_updated:
            LOG.debug("Tracking updates")
            self.add_column_mapping("_updated", dt.timestamp)
            self.add_index(["_updated"])
            self.__ignore_atts.append("_updated")

        if self.__track_created:
            LOG.debug("Tracking created timestamp")
            self.add_column_mapping("_created", dt.timestamp)
            self.add_index(["_created"])
            self.__ignore_atts.append("_created")

        if self.deleted_rows_method == "mark":
            self.add_column_mapping("_deleted", dt.timestamp)
            self.add_index(["_deleted"])
            self.__ignore_atts.append("_deleted")




import simpleetl._functions._datatypefuncs as _tfuncs
from simpleetl import LOG as _LOG

from simpleetl._modules.Datatype import Datatype as _Datatype

smallint = _Datatype("smallint", _tfuncs.is_smallint)
smallint_notnull = _Datatype("smallint", _tfuncs.is_smallint, False)

int = _Datatype("integer", _tfuncs.is_int)
int_notnull = _Datatype("integer", _tfuncs.is_int, False)

bigint = _Datatype("bigint", _tfuncs.is_bigint)
bigint_notnull = _Datatype("bigint", _tfuncs.is_bigint, False)

numeric7_2 = _Datatype("numeric(7,2)", _tfuncs.generate_is_numeric(7, 2))
numeric8_6 = _Datatype("numeric(8,6)", _tfuncs.generate_is_numeric(7, 2))
numeric3_1 = _Datatype("numeric(3,1)", _tfuncs.generate_is_numeric(3, 1))


def varchar(length):
    if length < 1:
        _LOG.error("varchar length must be at least 1")
    return _Datatype("varchar(" + str(length) + ")", _tfuncs.generate_is_varchar(length))


text = _Datatype("text", _tfuncs.is_text)
text_notnull = _Datatype("text", _tfuncs.is_text, True)
textlower = _Datatype("text", _tfuncs.is_textlower)
textlower_notnull = _Datatype("text", _tfuncs.is_textlower, True)
textupper = _Datatype("text", _tfuncs.is_textupper)
textupper_notnull = _Datatype("text", _tfuncs.is_textupper, True)

timekey = _Datatype("integer", _tfuncs.is_timekey)
time = _Datatype("time without time zone", _tfuncs.is_text)
timestamp = _Datatype("timestamp without time zone", _tfuncs.is_timestamp)
timestamp_with_timezone = _Datatype("timestamp with time zone", _tfuncs.is_timestamp)

boolean = _Datatype("boolean", _tfuncs.is_boolean)

datekey = _Datatype("integer", _tfuncs.is_datekey)
date = _Datatype("date", _tfuncs.is_text)

geography = _Datatype("geography", _tfuncs.is_text)

jsonb = _Datatype("jsonb", _tfuncs.injson)

xml = _Datatype("xml", _tfuncs.inxml)

numeric18_10 = _Datatype("numeric(18,10)", _tfuncs.generate_is_numeric(18, 10))

uuid = _Datatype("uuid", _tfuncs.is_uuid)
intarray = _Datatype("INTEGER[]", _tfuncs.is_intarray)
################################################################################
# spatial
################################################################################
# WGS-84
latitude = _Datatype('numeric(18,10)', _tfuncs.is_latitude)
# WGS-84
longitude = _Datatype('numeric(18,10)', _tfuncs.is_longitude)
# WGS-84
altitude = _Datatype('numeric(18,10)', _tfuncs.is_altitude)
# Compass direction/degree [0,360], assumed to be an integer
degree = _Datatype('int', _tfuncs.is_degree)

################################################################################
# XML Schema inspired basic data types
################################################################################
non_negative_int = _Datatype('int', _tfuncs.is_non_negative_int)
non_positive_int = _Datatype('int', _tfuncs.is_non_positive_int)
positive_int = _Datatype('int', _tfuncs.is_positive_int)



import logging
import os
import queue
import time
from multiprocessing import Queue, Event, Process

from simpleetl._modules.PostgresLoggingHandler import pgLogHandler


class LOGCLASS:
    def __init__(self):

        self.__formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s'
                                             ' - %(message)s')

        self.__mainpid = os.getpid()
        self.__stdlog = None
        self.__errlog = None
        self.__log_console = None
        self.__pglogger = None

        self.__logqueue = Queue(5)
        self.__logqueue.cancel_join_thread()  # Makes sure queue does not make main thread hang if main thread dies
        self.__running = Event()

        self.__log_lines = 0
        self.__log_mapper = {"debug": self.__debug,
                             "info": self.__info,
                             "warning": self.__warning,
                             "error": self.__error}
        # self.__logthread = threading.Thread(target=self.__logworker)
        self.__log_batchkey = -1
        self.__logthread = None

        self.__start_stdlog()

    def start_parallel(self):
        self.__logthread = Process(target=self.__logworker)
        self.__running.set()

        self.__logthread.start()

    def __logworker(self):
        from simpleetl._processing.ETLProcess import ETLProcess

        while True:
            try:
                os.kill(self.__mainpid, 0)  # Test if main process is alive. Throws PermissionError if not
                data = self.__logqueue.get(timeout=3)
            except queue.Empty:
                continue
            except EOFError:
                break
            except KeyboardInterrupt:
                break
            except BrokenPipeError:
                break
            except PermissionError:
                break
            if data == "STOP":
                break
            self.__log_lines += 1
            msgtype, msg = data
            self.__log_mapper[msgtype](msg)

            if self.__log_lines > 1000:
                self.__log_mapper["error"]("Error, more than 1000 lines logged. Something is terribly wrong!")
                ETLProcess.set_shutdown()
                break

        self.__running.clear()
        for handler in self.__stdlog.handlers:
            handler.flush()

    def set_batchkey(self, batchkey):
        self.__log_batchkey = batchkey

    def stop_parallel(self):
        self.__logqueue.put("STOP")
        while self.__running.is_set():
            time.sleep(0.1)

    def connect_batchlog(self, batchkey):
        if self.__pglogger:
            self.__pglogger.connect_batch(batchkey)

    def set_name(self, name):
        """

        :param name: Define name of logging instance

        """
        self.__stdlog.name = name

    # def set_debug(self):
    #    self.__log_console.setLevel(logging.DEBUG)

    def log_file(self, logpath, filename):
        # full_path = self.__logpath + "/logs/" + time.strftime("%Y_%m/%d/")
        os.makedirs(logpath, exist_ok=True)

        log_path = (logpath + "/" + filename)
        __log_file = logging.FileHandler(log_path)

        __log_file.setFormatter(self.__formatter)
        __log_file.setLevel(logging.DEBUG)
        self.__stdlog.addHandler(__log_file)

    def log_db(self, db_host, db_name, db_user, db_pass, db_port):
        self.__pglogger = pgLogHandler(self.__log_batchkey, db_host, db_name, db_user, db_pass, db_port)
        self.__stdlog.addHandler(self.__pglogger)

    def __start_stdlog(self):

        self.__stdlog = logging.getLogger("SimpleETL")
        self.__stdlog.setLevel(logging.DEBUG)

        self.__log_console = logging.StreamHandler()
        self.__log_console.setFormatter(self.__formatter)
        self.__log_console.setLevel(logging.DEBUG)
        self.__stdlog.addHandler(self.__log_console)

    def __start_errlog(self):
        logpath = self.__logpath + "/errlogs/" + time.strftime("%Y_%m/%d/")

        os.makedirs(logpath, exist_ok=True)
        log_path = (logpath + time.strftime('/%H_%M_%S_') + self.__logname + ".log")

        self.__errlog = logging.getLogger("BI ETL ERRORS")
        self.__errlog.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s'
                                      ' - %(message)s')

        __log_file = logging.FileHandler(log_path)
        __log_file.setFormatter(formatter)
        __log_file.setLevel(logging.DEBUG)
        self.__errlog.addHandler(__log_file)

    def __debug(self, message):
        self.__stdlog.debug(message)

    def __info(self, message):
        self.__stdlog.info(message)

    def __warning(self, message):
        self.__stdlog.warning(message)

    def __error(self, message):
        self.__stdlog.error(message)


    def __handle_log(self, message):
        if self.__running.is_set():
            self.__logqueue.put(message)
        else:
            msgtype, msg = message
            self.__log_mapper[msgtype](msg)

    def debug(self, message):
        self.__handle_log(("debug", message))

    def info(self, message):
        self.__handle_log(("info", message))

    def warning(self, message):
        self.__handle_log(("warning", message))

    def error(self, message, tbstr=None):
        self.__handle_log(("error", message))
        if tbstr:
            self.__handle_log(("error", "Stack trace:\n" + tbstr))



import os
import sys
from datetime import datetime
from multiprocessing import Event

import psycopg2
from pygrametl.tables import definequote

from simpleetl import LOG
from simpleetl._functions import _db_schema_handler as schand
from simpleetl._functions._dbfunctions import get_lock, release_lock
from simpleetl._modules.Batchlog import Batchlog
from simpleetl._processing.Postprocess import ETLPostprocess
from simpleetl._processing.Transform import ETLTransform


class ETLProcess(ETLTransform, ETLPostprocess):
    # __manager = Manager()
    __mp_error_event = Event()

    __mainpid = os.getpid()

    def __init__(self,
                 facttable, datafeeder,
                 db_host,
                 db_user, db_pass,
                 db_name, db_port,
                 data_timestamp=None,
                 prefunc=None,
                 processfunc=None):

        self.__facttable = facttable

        self.__db_dsn = ("dbname={dbname} user={username} password={password} host={host} port={port}"
                         .format(dbname=db_name, username=db_user, password=db_pass, host=db_host, port=db_port))
        # If no Type 2 SlowlyChangingFact date is set, use today

        if data_timestamp is None:
            self.__data_timestamp = datetime.utcnow()
        else:
            self.__data_timestamp = data_timestamp

        """ ---------- Initialize ---------- """

        definequote('"')
        self.__mp_error_event.clear()
        self.__batchlog = Batchlog(self.__db_dsn)
        self.__batchlog.new_batch(os.path.abspath(sys.argv[0]), " ".join(sys.argv[1:]),
                                  facttable.schema, facttable.table)
        self.__batchlog.set_start()
        LOG.connect_batchlog(self.__batchlog.get_batchkey())

        ETLTransform.__init__(self, facttable, datafeeder, self.__batchlog, self.__data_timestamp, self.__db_dsn,
                              prefunc, processfunc)
        ETLPostprocess.__init__(self, facttable, self.__batchlog, self.__data_timestamp, self.__db_dsn)

        self.__dbcon_lock = None

    @classmethod
    def set_shutdown(cls):
        cls.__mp_error_event.set()

    @classmethod
    def is_shutdown(cls):
        try:
            os.kill(cls.__mainpid, 0)  # Test if main process is alive. Throws PermissionError if not
        except PermissionError:
            cls.set_shutdown()
            return True
        return cls.__mp_error_event.is_set()

    def lock_tmpfacttable(self):
        if self.__dbcon_lock is None:
            self.__dbcon_lock = psycopg2.connect(self.__db_dsn,
                                                 application_name="Fact table lock" + self.__facttable.schema + "." +
                                                                  self.__facttable.table)
            self.__dbcon_lock.set_session(autocommit=True)
        get_lock(self.__dbcon_lock, self.__facttable.tmptable, False)

    def release_tmpfacttable(self):
        if self.__dbcon_lock is not None:
            release_lock(self.__dbcon_lock, self.__facttable.tmptable)
        self.__dbcon_lock.close()
        self.__dbcon_lock = None

    def lock_facttable(self):
        if self.__dbcon_lock is None:
            self.__dbcon_lock = psycopg2.connect(self.__db_dsn,
                                                 application_name="Fact table lock" + self.__facttable.schema + "." +
                                                                  self.__facttable.table)
            self.__dbcon_lock.set_session(autocommit=True)
        get_lock(self.__dbcon_lock, self.__facttable.schema + "." +
                 self.__facttable.table, False)

    def release_facttable(self):
        if self.__dbcon_lock is not None:
            release_lock(self.__dbcon_lock, self.__facttable.schema + "." + self.__facttable.table)

    def ensure_structure(self):
        # gotlock = dbcon.lock_table(facttable.schema, facttable.table, True)

        # if not gotlock:
        #    LOG.debug("Fact table {s}.{t} is currently utilized by other ETL process. Waiting.."
        #              .format(s=facttable.schema, t=facttable.table))
        #    dbcon.lock_table(facttable.schema, facttable.table, False)

        self.__facttable._validate_structure()
        self.__facttable._prepare_run()

        LOG.info("-------------------- " +
                 "{s}.{t}".format(s=self.__facttable.schema, t=self.__facttable.table) +
                 " --------------------")
        LOG.info("Starting ETL for {s}.{t}".format(s=self.__facttable.schema, t=self.__facttable.table))
        LOG.info("Command: " + " ".join(sys.argv[0:5]) +
                 (len(sys.argv) > 5 and " [more..]" or ""))

        self.__batchlog.set_wait_start()

        pgconn = psycopg2.connect(self.__db_dsn, application_name="Create schema for " +
                                                                  "{s}.{t}".format(s=self.__facttable.schema,
                                                                                   t=self.__facttable.table))

        get_lock(pgconn, "ETL create_schema")
        for dim in self.__facttable.dim_mappings:
            dim.dimension.create_schema(pgconn)

        # Ensure fact table structure

        schand._ensure_structure(pgconn, self.__facttable)
        schand._ensure_indexes(pgconn, self.__facttable)

        # Create table for storing temporary data.

        schand._create_temp_table(pgconn, self.__facttable)

        pgconn.commit()
        release_lock(pgconn, "ETL create_schema")
        pgconn.commit()
        """ ---------- Extract/Transform ---------- """
        pgconn.close()

    def status(self):
        return self.__batchlog.get_stats()



from simpleetl._functions._db_quoting import quote


def _get_nextid(pgconn, schema, table, key):
    q = """SELECT
        coalesce(MAX({key})+1,1) as maxid
        FROM {schema}.{table};""".format(schema=quote(schema, pgconn),
                                         table=quote(table, pgconn),
                                         key=quote(key, pgconn))
    with pgconn.cursor() as cursor:
        cursor.execute(q)
        return cursor.fetchone()[0]



from pygrametl.tables import BulkFactTable, _quote
import pygrametl


def _get_bulktable(pygramcon, schema, table, keyrefs, measures, incl_id=None):
    # keyrefs = self.get_cols(atts=False)
    # measures = self.get_cols(dims=False)

    if incl_id is not None:
        measures.append(incl_id)

    facttable = BulkFactTable(name=_quote(schema) + "." + _quote(table),
                              keyrefs=keyrefs,
                              measures=measures,
                              bulkloader=generate_bulkloader(pygramcon),
                              fieldsep='\t',
                              rowsep='\n',
                              nullsubst="\\N",
                              tempdest=None,
                              bulksize=100000,
                              usefilename=False,
                              dependson=(),
                              strconverter=getdbfriendlystr)

    return facttable


def _get_bulkhisttable(pygramcon, schema, table, table_historic, keyrefs, measures, key, incl_id=None):
    measures += ['_validfrom', '_validto', '_version']
    if incl_id is not None:
        measures.append(incl_id)

    facttable = BulkFactTable(name=_quote(schema) + "." + _quote(table_historic),
                              keyrefs=keyrefs,
                              measures=measures,
                              bulkloader=generate_bulkloader(pygramcon),
                              fieldsep='\t',
                              rowsep='\n',
                              nullsubst="\\N",
                              tempdest=None,
                              bulksize=100000,
                              usefilename=False,
                              dependson=(),
                              strconverter=getdbfriendlystr)
    return facttable


def generate_bulkloader(pygramcon):
    def pgcopybulkloader(name, atts, fieldsep, rowsep, nullval,
                         filehandle):
        """ A bulk loader method for loading data into the data warehouse.

        This is an interface between pygrametl and psycopg2, hence the argument
        list is required for pygrametl to be working.

        :param name: Table name.
        :param atts: Column names to insert into.
        :param fieldsep: Seperator of the CSV file.
        :param rowset: Row seperator of the CSV file. Not used by psycopg2, but
            required for pygrametl to be working.
        :param nullval: Null value.
        :param filehandler: A file handler, from where the data is loaded from.
        """

        with pygramcon.cursor() as cursor:
            cursor.copy_from(file=filehandle, table=name, sep=fieldsep,
                             null=str(nullval),
                             columns=[pygrametl.tables._quote(xn) for xn in atts])

    return pgcopybulkloader


def getdbfriendlystr(value, nullvalue='None'):
    """Covert a value into a string that can be accepted by a DBMS.

       None values are converted into the value of the argument nullvalues
       (default: 'None'). Bools are converted into '1' or '0' (instead of
       'True' or 'False' as str would do). Other values are currently just
       converted by means of str.

    """
    if value is None:
        return nullvalue
    elif isinstance(value, bool):
        if value:
            return "1"
        else:
            return "0"
    elif isinstance(value, str):
        return value.replace("\n", "\\n").replace("\t","\\t").replace("\r","\\r")
    else:
        return str(value)



import calendar
from datetime import datetime, date

import pg_creds
from simpleetl import CONFIG
from simpleetl import FactTable
from simpleetl import LOG
from simpleetl import datatypes as dt
from simpleetl import runETL
from examples.datedimension_example import dimdate
from examples.timedimension_example import dimtime


def data_generator():
    """ Simple data generator, yielding dictionaries of data.
    Could as well be a file parser or anything else
    :return:
    """
    userdata = [{"userid": 42, "sys_username": "Jens",
                 "created": datetime.strptime("2019-02-26 07:32:15", "%Y-%m-%d %H:%M:%S")},
                {"userid": 56, "sys_username": "Svend",
                 "created": datetime.strptime("2019-04-11 12:54:11", "%Y-%m-%d %H:%M:%S")},
                {"userid": 78, "sys_username": "Niels",
                 "created": datetime.strptime("2019-07-21 07:12:52", "%Y-%m-%d %H:%M:%S")}
                ]
    for user in userdata:
        yield user


def partitionfunc_yearmonth(value):
    """Function for telling simpleetl to partition facttable by year and month.
    :param value: Input is an integer datekey, ex. '2019-11-23' is 20191123
    :return: Tuple with suffix, range_from and range_to, ex. ('201911', 20191101, 20191131)
    """
    dt = datetime.strptime(str(value), "%Y%m%d")
    first_day, last_day = calendar.monthrange(dt.year, dt.month)
    suffix = dt.strftime("%Y%m")  # 201911
    range_from = int(date(dt.year, dt.month, 1).strftime("%Y%m%d"))  # 20191101, ensures zeropadded month
    range_to = int(date(dt.year, dt.month, last_day).strftime("%Y%m") + str(last_day + 1))  # 20191131
    return suffix, range_from, range_to


def processrow(row):
    """ Function for transforming each single row.
    :param row: Input data row as dictionary. Data transformations must be updated in this dictionary,
        e.g., row["name_lower"] = row["name"].lower()
    :return: Can return True or None if row is fine. If function returns false, row will be discarded and not inserted.
    """
    if row["created"] is None:
        # If we have no created timestamp, row is invalid and we discard it
        return False

    row["created_datekey"] = int(row["created"].strftime("%Y%m%d"))
    # The date dimension expects a datekey smart key as integer

    row["created_timekey"] = int(row["created"].strftime("%H%M"))
    # The time dimension expects a timekey smart key as integer


def load_user_etl():
    LOG.info("Starting to define ETL")
    p = FactTable(schema="testschema", table="userdata",
                  migrate_updates=True,
                  # Updated to data will be processed. Can be set to False if only appending (will speed things up)
                  store_history=False,  # Create a seperate userdata_historic table for storing changes to facts.
                  track_last_updated=True,
                  # Adds an _updated attribute which keeps track of when data was last updated.
                  lookupatts=["userid", "created_datekey"]  # List of attributes uniquely defining a fact.
                  # created_datekey is required in lookupatts, as the table will be partitioned by this attribute.
                  )

    p.enable_partitioning("created_datekey", partitionfunc_yearmonth)
    # Enable partitioning on the table. The table will be partitioned by the created_datekey attribute
    # and partition will be found by using the partitionfunc_yearmonth

    p.handle_deleted_rows("mark")
    # Tells ETL to mark deleted rows from source with an _deleted timestamp attribute

    p.add_column_mapping("userid", dt.bigint, "userid")
    # Map userid from source data to database with same name

    p.add_column_mapping("sys_username", dt.text, "username")
    # Rename "sys_username" from source data to "username" in database

    p.add_column_mapping("created", dt.timestamp, "created_timestamp")
    # Remap "created" in source data to "created_timestamp" in database

    p.add_dim_mapping(dimdate, "created_datekey", {"datekey": "created_datekey"})
    # dimdate expects a datekey, hence we create a namemapping from created_datekey to datekey is created

    p.add_dim_mapping(dimtime, "created_timekey", {"timekey": "created_timekey"})
    # dimtime expects a timekey attribyte, hence a namemapping from created_timekey to timekey is created

    datafeeder = data_generator()
    # datafeeder is our data generator object
    #

    CONFIG.print_progress_line = 1
    # Make simpleETL print progress for every n lines

    CONFIG.debug = False
    # Enabled debugging of ETL. Will not drop temporary and will print SQLs for easier debugging.

    # CONFIG.simulate = False # Can simulate effects of ETL run
    # CONFIG.batch_size = 1000 Batch size of rows being distributed for each parallel worker
    # CONFIG.parallel_processes = multiprocessing.cpu_count() #Increase or decrease number of parallel bach workers.
    # CONFIG.batch_logging_table = True # Simpleetl will log batchload metadata to simpleetl.batchload table

    LOG.info("Executing ETL process")
    stats = runETL(facttable=p,
                   datafeeder=datafeeder,
                   prefunc=None,
                   # A function for each row running in the main thread can be defined. Can be usefull,
                   # e.g., if it is required to compare data globaly to detect duplicates.
                   processfunc=processrow,  # Each row must be processed by this function in each worker.
                   db_host=pg_creds.PG_HOST, db_port=pg_creds.PG_PORT,
                   db_name=pg_creds.PG_DBNAME, db_user=pg_creds.PG_USER,
                   db_pass=pg_creds.PG_PASS)
    print(stats)
    return stats


if __name__ == "__main__":
    load_user_etl()



class Datatype():
    def __init__(self, sqltype, tfunc, allow_null=True):
        self.__sqltype = sqltype
        self.__tfunc = tfunc
        self.__allow_null = allow_null

    def sqltype(self):
        if self.__allow_null:
            return self.__sqltype
        else:
            return self.__sqltype + " NOT NULL"

    def parse(self, val):
        return self.__tfunc(val)

    def copy(self, allow_null):
        return Datatype(self.__sqltype, self.__tfunc, allow_null)

    @property
    def allow_null(self):
        return self.__allow_null



import hashlib
from functools import lru_cache

from psycopg2.extensions import quote_ident


def table_exists(pgconn, schema, table):
    """ Returns true, if table exists in data warehouse.
    :param pgconn: Psycopg2 connection
    :param schema: Schema name.
    :param table: Table name.
    :return: True if table exists, else False.
    """

    stmt = "SELECT COUNT(1) AS n FROM pg_tables WHERE tablename=%s AND schemaname=%s"
    with pgconn.cursor() as cursor:
        cursor.execute(stmt, [table, schema])
        return cursor.fetchone()[0] == 1


def has_index(pgconn, schema, table, cols):
    cols2 = cols.copy()
    for col in cols:
        cols2.append(quote_ident(col, pgconn))
    q = """with idxs as (SELECT i.relname as indname,
       i.relowner as indowner,
       idx.indrelid::regclass,
       am.amname as indam,
       idx.indkey,
       ARRAY(
       SELECT replace(replace(pg_get_indexdef(idx.indexrelid, k + 1, true),'"',''),'\\','')
       FROM generate_subscripts(idx.indkey, 1) as k
       ORDER BY k
       ) as indkey_names,
       idx.indexprs IS NOT NULL as indexprs,
       idx.indpred IS NOT NULL as indpred
FROM   pg_index as idx
JOIN   pg_class as i
ON     i.oid = idx.indexrelid
JOIN   pg_am as am
ON     i.relam = am.oid
where idx.indrelid::regclass ='{schema}.{table}'::regclass)
select * from idxs
    where indkey_names = %s
--
    """.format(schema=quote_ident(schema, pgconn),
               table=quote_ident(table, pgconn))
    with pgconn.cursor() as cursor:
        cursor.execute(q, (cols,))

    return cursor.rowcount > 0


def column_exists(pgconn, schema, table, column):
    """ Returns true, if column in table exists in data warehouse.
    :param pgconn: psycopg2 connection.
    :param schema: Schema name.
    :param table: Table name.
    :param column: Table column
    :return: True if table exists, else False.
    """

    sql = """
    SELECT
        COUNT(1) AS n
    FROM information_schema.columns
    WHERE table_schema = %s
        AND table_name = %s
        AND column_name = %s;
        """
    with pgconn.cursor() as cursor:
        cursor.execute(sql, (schema, table, column))
        tablecount = cursor.fetchone()
    return tablecount[0] == 1


def column_type(pgconn, schema, table, column):
    """ Returns datatype, if column in table exists, otherwise returns None.
        :param pgconn: psycopg2 connection.
        :param schema: Schema name.
        :param table: Table name.
        :param column: Table column
        :return: Column if exists, otherwise None.
        """

    q_ddl = """	SELECT udt_name::regtype as datatype,
                is_nullable,
                numeric_precision,
                numeric_scale
            FROM INFORMATION_SCHEMA.COLUMNS WHERE table_schema=%s 
            AND table_name=%s
            AND column_name=%s
            """
    with pgconn.cursor() as cursor:
        cursor.execute(q_ddl, (schema, table, column))

        if cursor.rowcount == 0:
            return None
        col = cursor.fetchone()

    data_type = col[0]
    is_nullable = col[1]
    numeric_precision = col[2]
    numeric_scale = col[3]

    if data_type == "numeric":
        data_type += "({p},{s})".format(p=numeric_precision, s=numeric_scale)
    if is_nullable == "NO":
        data_type += " NOT NULL"
    return data_type


def has_fkey(pgconn, srcschema, srctable, srccol, dstschema, dsttable, dstcol):
    q = """select 1
    from pg_attribute af, pg_attribute a,
      (select conrelid,confrelid,conkey[i] as conkey, confkey[i] as confkey
       from (select conrelid,confrelid,conkey,confkey,
                    generate_series(1,array_upper(conkey,1)) as i
             from pg_constraint where contype = 'f') ss) ss2
    where af.attnum = confkey and af.attrelid = confrelid and
          a.attnum = conkey and a.attrelid = conrelid

          and confrelid::regclass='{dstschema}.{dsttable}'::regclass
          and af.attname ='{dstcol}'
          and conrelid::regclass = '{srcschema}.{srctable}'::regclass
          and a.attname ='{srccol}'
          --and replace(confrelid::regclass::text,'"','')=replace('{dsttable}','"','')
          --and replace(af.attname,'"','') =replace('{dstcol}','"','')
          --and replace(conrelid::regclass::text,'"','') = replace('{srctable}','"','')
          --and replace(a.attname,'"','') =replace('{srccol}','"','')
    """.format(srcschema=quote_ident(srcschema, pgconn),
               srctable=quote_ident(srctable, pgconn),
               srccol=srccol,
               dstschema=quote_ident(dstschema, pgconn),
               dsttable=quote_ident(dsttable, pgconn),
               dstcol=dstcol)
    with pgconn.cursor() as cursor:
        cursor.execute(q)

    return cursor.rowcount > 0


def has_unique(pgconn, schema, table, cols):
    q = """select
    b.relname as table,
    a.conname as constraint,
    array_agg(c.attname) as const_columns
from pg_constraint a, pg_class b, pg_attribute c
where
	contype='u' and 
 conrelid::regclass=%s::regclass
    and a.conrelid = b.oid
    and c.attrelid = b.oid
    and c.attnum in (select unnest(a.conkey))
group by b.relname, a.conname"""

    with pgconn.cursor() as cursor:
        cursor.execute(q, (schema + "." + table,))
        const = cursor.fetchall()

    for c in const:
        constlist = c[2]
        if set(constlist) == set(cols):
            return True
    return False


def create_unique(pgconn, schema, table, cols):
    q = """ALTER TABLE {schema}.{table} ADD UNIQUE ({cols})
    """.format(schema=quote_ident(schema, pgconn),
               table=quote_ident(table, pgconn),
               cols=", ".join([quote_ident(c, pgconn) for c in cols]))
    with pgconn.cursor() as cursor:
        cursor.execute(q)


def create_fkey(pgconn, srcschema, srctable, srccol, dstschema, dsttable, dstcol, deferrable=False):
    defer = ""
    if deferrable:
        defer = "deferrable initially deferred"
    q = """alter table {srcschema}.{srctable} add foreign key ({srccol}) references {dstschema}.{dsttable} ({dstcol}) {defer}
    """.format(srcschema=quote_ident(srcschema, pgconn),
               srctable=quote_ident(srctable, pgconn),
               srccol=quote_ident(srccol, pgconn),
               dstschema=quote_ident(dstschema, pgconn),
               dsttable=quote_ident(dsttable, pgconn),
               dstcol=quote_ident(dstcol, pgconn),
               defer=defer)
    with pgconn.cursor() as cursor:
        cursor.execute(q)


@lru_cache(100)
def __parse_lockname(name):
    return int(hashlib.md5(name.encode()).hexdigest(), 16) % (2 ** 63)


def get_lock(pgconn, name, dotry=False):
    getlock = "pg_advisory_lock"
    name = __parse_lockname(name)
    if dotry:
        getlock = "pg_try_advisory_lock"
    with pgconn.cursor() as cursor:
        cursor.execute("SELECT {getlock}({name}) as l"
                       .format(getlock=getlock, name=name))
        if dotry:
            r = cursor.fetchone()
        else:
            return True

    return r[0]


def get_lock_transaction(pgconn, name, dotry=False):
    """Sets an transaction lock with the given name. Transaction lock is released upon commit or rollback.
    :param pgconn: Postgresql connection
    :param name: Name of lock
    param dotry: If True, tries to fetch lock and returns whether it was possible. If False, default,
        wait until lock is available.
    """
    getlock = "pg_advisory_xact_lock"
    name = __parse_lockname(name)
    if dotry:
        getlock = "pg_try_advisory_xact_lock"

    with pgconn.cursor() as cursor:
        cursor.execute("SELECT {getlock}({name}) as l"
                       .format(getlock=getlock, name=name))
        if dotry:
            r = cursor.fetchone()
        else:
            return True

    return r[0]


def release_lock(pgconn, name):
    name = __parse_lockname(name)
    unlock = "pg_advisory_unlock"
    with pgconn.cursor() as cursor:
        cursor.execute("SELECT {unlock}({name}) as l"
                       .format(unlock=unlock, name=name))
        r = cursor.fetchone()
    return r[0]


def get_table_partitions(pgconn, schema, table):
    get_partitions = """with recursive inh as (
       select i.inhrelid, null::text as parent
       from pg_catalog.pg_inherits i
         join pg_catalog.pg_class cl on i.inhparent = cl.oid
         join pg_catalog.pg_namespace nsp on cl.relnamespace = nsp.oid
       where nsp.nspname = '{s}' and cl.relname = '{t}'
       union all
       select i.inhrelid, (i.inhparent::regclass)::text
       from inh
         join pg_catalog.pg_inherits i on (inh.inhrelid = i.inhparent)
    )
    select c.relname as partition_name
    from inh
       join pg_catalog.pg_class c on inh.inhrelid = c.oid
       ORDER BY c.relname""".format(s=schema, t=table)
    output = []
    with pgconn.cursor() as cursor:
        cursor.execute(get_partitions)
        for p in cursor.fetchall():
            output.append(p[0])
    return output


def is_table_partitioned(pgconn, schema, table):
    get_partitions = """SELECT count(1) > 0 as exist 
FROM pg_partitioned_table 
WHERE partrelid::regclass::text=%s"""

    r = False
    with pgconn.cursor() as cursor:
        cursor.execute(get_partitions, (schema + "." + table,))
        r = cursor.fetchone()[0]
    return r



import sys
import time
import traceback

from simpleetl import CONFIG
from simpleetl import LOG
from simpleetl._processing.ETLProcess import ETLProcess

__all__ = ['runETL']


def runETL(facttable, datafeeder,
           db_host,
           db_user, db_pass,
           db_name, db_port,
           data_timestamp=None,
           prefunc=None,
           processfunc=None):
    """ Perform ETL processing.
        :param facttable: Object containing information on Fact Tables.
        :param datafeeder: A data feeder iterator, returning dictionaries.
        :param db_host: Database host.
        :param db_user: Database user.
        :param db_pass: Database pass.
        :param db_name: Database name.
        :param db_port: Optionally, a database port. Defaults to 5432.
        :param data_timestamp: A timestamp defining when a fact is valid. If None, the timestamp for starting ETL will be used.
            This parameter is only used when having history tracking enabled.
        :param prefunc: A "pre-work" function, func(row),that will be executed before splitting data into parallel processing.
            Must return True if rows is valid, otherwise False, which ignores the row.
        :param processfunc: A "parallel-work" function func(row,p_dbcon), that will be executed in parallel, before generic data processing
            If False is returned from function, row will be discarded.
        :return: Dictionary with status data
        """
    t = time.time()
    if CONFIG.debug:
        LOG.info(
            "========== Running ETL in DEBUG mode. Debug messages will be printed and errors not handled ==========")
    e = ETLProcess(facttable, datafeeder,
                   db_host,
                   db_user, db_pass,
                   db_name, db_port,
                   data_timestamp=data_timestamp,
                   prefunc=prefunc,
                   processfunc=processfunc)
    e.lock_tmpfacttable()
    e.ensure_structure()
    try:
        LOG.start_parallel()
        success = e.start_etl()
        if CONFIG.debug: LOG.debug("ETL method completed with status " + str(success))
    except Exception as err:
        tb = traceback.format_exc()
        errmsg = "ETL"

        LOG.error(errmsg, "".join(tb))
        success = False
    if CONFIG.debug: LOG.debug("Stopping logger")

    if success:
        if CONFIG.debug: LOG.debug("Final load")
        e.lock_facttable()
        success = e.final_load()
        e.release_facttable()

    e.release_tmpfacttable()

    LOG.stop_parallel()
    if not success:
        tsstr = time.strftime("%H:%M:%S", time.gmtime(time.time() - t))
        LOG.error("ETL JOB FAILED in " + tsstr)
        if CONFIG.debug: LOG.error("sys.exit(1)")
        sys.exit(1)
    else:
        tsstr = time.strftime("%H:%M:%S", time.gmtime(time.time() - t))
        LOG.info("ETL JOB completed in " + tsstr)
    return e.status()